from typing import Literal
from jaxtyping import Shaped, Float, Inexact
from torch import Tensor

from functools import partial
from math import prod, ceil

import torch

try:
    import triton
    import triton.language as tl

    TRITON_ENABLED = True
except ImportError:
    from torchlinops.utils import fake_triton as triton, fake_tl as tl

    TRITON_ENABLED = False

from .kernels import (
    weights1d,
    weights2d,
    weights3d,
    weights_torch,
    get_kernel_fn,
    _apply_default_kernel_params,
    KernelTypeStr,
    mod_pos,
)
from ._batch import batch_iterator

__all__ = ["grid"]

# Limit maximum number of grids to launch
TRITON_MAX_GRID_SIZE = 2**16 - 1
# Limit maximum kernel width in each dimension
TRITON_MAX_KERNEL_WIDTH_1D = 2**6


def grid(
    vals: Inexact[Tensor, "..."],
    locs: Float[Tensor, "... D"],
    grid_size: tuple[int, ...],
    width: float | tuple[float, ...],
    kernel: str = "kaiser_bessel",
    norm: str = "1",
    pad_mode: Literal["zero", "circular"] = "circular",
    kernel_params: dict = None,
):
    """Interpolate from off-grid values to on-grid locations.

    grid_size : tuple[int, ...]
        Shape of output array, excluding batch dimensions
        For example, if the gridded output should have shape [3, 64, 64],
        where 3 is the batch size, then grid_size would be [64, 64]

    norm: str, 1 or 2
        if 2, uses Euclidean norm to grid points to compute weights
        if 1, computes weights as product of axis-aligned norm weights
            - Same as sigpy
    """
    kernel_params = {} if kernel_params is None else kernel_params
    vals_flat, locs, shapes = prep_grid_shapes(vals, locs, grid_size, width)
    kernel_params = _apply_default_kernel_params(kernel, kernel_params)
    out_flat = _grid(
        vals_flat,
        locs,
        kernel=kernel,
        norm=norm,
        pad_mode=pad_mode,
        kernel_params=kernel_params,
        **shapes,
    )
    out = out_flat.reshape(*shapes["batch_shape"], *grid_size)
    return out


def _grid(
    vals: Shaped[Tensor, "B ..."],
    locs: Float[Tensor, "... D"],
    grid_size: tuple[int, ...],
    width: tuple[float, ...],
    kernel: KernelTypeStr,
    norm: str,
    pad_mode: str,
    ndim: int,
    nbatch: int,
    is_complex: bool,
    locs_batch_shape: tuple[int, ...],
    npts: int,
    kernel_params,
    **kwargs,
):
    if vals.is_cuda and ndim in GRID.keys():
        # Ensure contiguity
        vals = vals.contiguous()
        locs = locs.contiguous()
        # Preallocate output
        output = torch.zeros(
            nbatch,
            *grid_size,
            dtype=vals.dtype,
            device=vals.device,
        )
        if is_complex:
            vals = torch.view_as_real(vals).contiguous()
            output = torch.view_as_real(output).contiguous()
        grid = _get_grid()
        BLOCK_WIDTH = get_block_width(width, ndim, is_complex)
        with torch.cuda.device(vals.device):
            GRID[ndim][grid](
                vals,
                locs,
                output,
                nbatch,
                npts,
                kernel,
                norm,
                pad_mode,
                is_complex,
                *grid_size,
                *width,
                *BLOCK_WIDTH,
                **kernel_params,
            )
        if is_complex:
            output = torch.view_as_complex(output)
    else:
        # Pytorch fallback
        output = grid_torch(
            vals,
            locs,
            grid_size,
            width,
            kernel,
            norm,
            pad_mode,
            **kernel_params,
        )
    return output


def _get_grid():
    grid = lambda meta: (ceil(meta["npts"] / meta["pts_per_grid"]) * meta["nbatch"],)  # noqa: E731
    return grid


def get_block_width(kernel_width: tuple[float, ...], ndim: int, is_complex: bool):
    """Get necessary block width based on dimension and dtype of input"""
    block_width = list(triton.next_power_of_2(ceil(w + 1)) for w in kernel_width)
    test_block_width = block_width[:]  # Shallow copy
    if is_complex:
        test_block_width[-1] *= 2
    if max(test_block_width) > TRITON_MAX_KERNEL_WIDTH_1D:
        raise ValueError(
            f"Necessary block width {test_block_width} has entry which exceeds maximum width {TRITON_MAX_KERNEL_WIDTH_1D} (kernel width is doubled in last dim if input is complex)"
        )

    return block_width


@triton.heuristics(
    values={
        "pts_per_grid": lambda args: max(
            1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
        ),
    },
)
@triton.jit
def _grid1d(
    in_ptr,
    pts_ptr,
    out_ptr,
    nbatch,
    npts,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    PAD_MODE: tl.constexpr,
    is_complex: tl.constexpr,  # bool
    x_size,
    x_kernel_width,
    X_BLOCK_WIDTH: tl.constexpr,
    pts_per_grid,  # Determined via heuristic
    beta=1.0,  # For kernel=kaiser_bessel
):
    """
    NORM has no effect in 1d
    """
    size = x_size

    pid_0 = tl.program_id(0)
    grids_per_batch = tl.cast(tl.ceil(npts / pts_per_grid), tl.int32)
    N, grid_start = pid_0 // grids_per_batch, pid_0 % grids_per_batch
    pts_lower, pts_upper = pts_per_grid * grid_start, pts_per_grid * (grid_start + 1)

    x_base_range = tl.arange(0, X_BLOCK_WIDTH)
    if is_complex:
        # Last dimension has double size because (real, imag) are interleaved
        # Overall size doubles as a result
        size = 2 * size

    in_batch_offset = N * npts
    out_batch_offset = N * size

    for p in range(pts_lower, pts_upper):
        if p < npts:
            x_target = tl.load(pts_ptr + p)
            weights, x_range, x_mask = weights1d(
                x_target, x_kernel_width, x_base_range, KERNEL, beta
            )

            if is_complex:
                # Pytorch interleaved indexing
                # Only applies to last dimension
                x_range_real = 2 * x_range
                x_range_imag = 2 * x_range + 1
                x_range_cplx = tl.join(x_range_real, x_range_imag)  # [width, 2]
                x_mask_cplx = tl.join(x_mask, x_mask)
                if PAD_MODE == "zero":
                    x_mask_cplx &= (x_range_cplx >= 0) & (x_range_cplx < (2 * x_size))
                elif PAD_MODE == "circular":
                    x_range_cplx = mod_pos(x_range_cplx, 2 * x_size)

                # Split and process separately
                pt_real = tl.load(in_ptr + 2 * (in_batch_offset + p))
                pt_imag = tl.load(in_ptr + 2 * (in_batch_offset + p) + 1)
                out_real = weights * pt_real
                out_imag = weights * pt_imag
                out_cplx = tl.join(out_real, out_imag)

                # Store
                tl.atomic_add(
                    out_ptr + out_batch_offset + x_range_cplx, out_cplx, x_mask_cplx
                )

            else:
                # Normal indexing
                if PAD_MODE == "zero":
                    x_mask &= (x_range >= 0) & (x_range < x_size)
                elif PAD_MODE == "circular":
                    x_range = mod_pos(x_range, x_size)

                # Load
                pt_val = tl.load(in_ptr + in_batch_offset + p)
                out = weights * pt_val

                # Store
                tl.atomic_add(out_ptr + out_batch_offset + x_range, out, x_mask)


@triton.heuristics(
    values={
        "pts_per_grid": lambda args: max(
            1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
        ),
    },
)
@triton.jit
def _grid2d(
    in_ptr,
    pts_ptr,
    out_ptr,
    nbatch,
    npts,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    PAD_MODE: tl.constexpr,
    is_complex: tl.constexpr,  # bool
    # Size of grid
    x_size,
    y_size,
    # Size of kernel
    x_kernel_width,
    y_kernel_width,
    # Size of blocks to load
    X_BLOCK_WIDTH: tl.constexpr,
    Y_BLOCK_WIDTH: tl.constexpr,
    pts_per_grid,  # Determined via heuristic
    beta=1.0,  # For kernel=kaiser_bessel
):
    """ """
    size = x_size * y_size

    pid_0 = tl.program_id(0)
    grids_per_batch = tl.cast(tl.ceil(npts / pts_per_grid), tl.int32)
    N, grid_start = pid_0 // grids_per_batch, pid_0 % grids_per_batch
    pts_lower, pts_upper = pts_per_grid * grid_start, pts_per_grid * (grid_start + 1)

    x_base_range = tl.arange(0, X_BLOCK_WIDTH)
    y_base_range = tl.arange(0, Y_BLOCK_WIDTH)
    if is_complex:
        # Last dimension has double size because (real, imag) are interleaved
        # Overall size doubles as a result
        size = 2 * size

    in_batch_offset = N * npts
    out_batch_offset = N * size

    for p in range(pts_lower, pts_upper):
        if p < npts:
            # Load target point
            x_target = tl.load(pts_ptr + 2 * p)
            y_target = tl.load(pts_ptr + 2 * p + 1)

            weights, x_range, y_range, x_mask, y_mask = weights2d(
                x_target,
                y_target,
                x_kernel_width,
                y_kernel_width,
                x_base_range,
                y_base_range,
                KERNEL,
                NORM,
                beta,
            )

            if is_complex:
                x_range_cplx = x_range  # [width]
                x_mask_cplx = x_mask
                # Pytorch interleaved indexing
                # Only applies to last dimension
                y_range_real = 2 * y_range
                y_range_imag = 2 * y_range + 1
                y_range_cplx = tl.join(y_range_real, y_range_imag)  # [width, 2]
                y_mask_cplx = tl.join(y_mask, y_mask)
                if PAD_MODE == "zero":
                    x_mask_cplx &= (x_range_cplx >= 0) & (x_range_cplx < x_size)
                    y_mask_cplx &= (y_range_cplx >= 0) & (y_range_cplx < (2 * y_size))
                elif PAD_MODE == "circular":
                    x_range_cplx = mod_pos(x_range_cplx, x_size)
                    y_range_cplx = mod_pos(y_range_cplx, 2 * y_size)

                grid_range_cplx = (
                    x_range_cplx[:, None, None] * y_size * 2 + y_range_cplx[None, :, :]
                )
                grid_mask_cplx = x_mask_cplx[:, None, None] & y_mask_cplx[None, :, :]

                # Load
                # Split and process separately
                pt_real = tl.load(in_ptr + 2 * (in_batch_offset + p))
                pt_imag = tl.load(in_ptr + 2 * (in_batch_offset + p) + 1)
                out_real = weights * pt_real
                out_imag = weights * pt_imag
                out_cplx = tl.join(out_real, out_imag)

                # Accumulate
                tl.atomic_add(
                    out_ptr + out_batch_offset + grid_range_cplx,
                    out_cplx,
                    grid_mask_cplx,
                )

            else:
                # Normal indexing
                if PAD_MODE == "zero":
                    x_mask &= (x_range >= 0) & (x_range < x_size)
                    y_mask &= (y_range >= 0) & (y_range < y_size)
                elif PAD_MODE == "circular":
                    x_range = mod_pos(x_range, x_size)
                    y_range = mod_pos(y_range, y_size)

                grid_range = x_range[:, None] * y_size + y_range[None, :]
                grid_mask = x_mask[:, None] & y_mask[None, :]

                # Load
                pt_val = tl.load(in_ptr + in_batch_offset + p)
                out = weights * pt_val

                # Accumulate
                tl.atomic_add(out_ptr + out_batch_offset + grid_range, out, grid_mask)


@triton.heuristics(
    values={
        "pts_per_grid": lambda args: max(
            1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
        ),
    },
)
@triton.jit
def _grid3d(
    in_ptr,
    pts_ptr,
    out_ptr,
    nbatch,
    npts,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    PAD_MODE: tl.constexpr,
    is_complex: tl.constexpr,  # bool
    # Size of grid
    x_size,
    y_size,
    z_size,
    # Size of kernel
    x_kernel_width,
    y_kernel_width,
    z_kernel_width,
    # Size of blocks to load
    X_BLOCK_WIDTH: tl.constexpr,
    Y_BLOCK_WIDTH: tl.constexpr,
    Z_BLOCK_WIDTH: tl.constexpr,
    pts_per_grid,  # Determined via heuristic
    beta=1.0,  # For kernel=kaiser_bessel
):
    """ """
    size = x_size * y_size * z_size

    pid_0 = tl.program_id(0)
    grids_per_batch = tl.cast(tl.ceil(npts / pts_per_grid), tl.int32)
    N, grid_start = pid_0 // grids_per_batch, pid_0 % grids_per_batch
    pts_lower, pts_upper = pts_per_grid * grid_start, pts_per_grid * (grid_start + 1)

    x_base_range = tl.arange(0, X_BLOCK_WIDTH)
    y_base_range = tl.arange(0, Y_BLOCK_WIDTH)
    z_base_range = tl.arange(0, Z_BLOCK_WIDTH)
    if is_complex:
        # Last dimension has double size because (real, imag) are interleaved
        # Overall size doubles as a result
        size = 2 * size

    in_batch_offset = N * npts
    out_batch_offset = N * size

    for p in range(pts_lower, pts_upper):
        if p < npts:
            # Load target point
            x_target = tl.load(pts_ptr + 3 * p)
            y_target = tl.load(pts_ptr + 3 * p + 1)
            z_target = tl.load(pts_ptr + 3 * p + 2)

            weights, x_range, y_range, z_range, x_mask, y_mask, z_mask = weights3d(
                x_target,
                y_target,
                z_target,
                x_kernel_width,
                y_kernel_width,
                z_kernel_width,
                x_base_range,
                y_base_range,
                z_base_range,
                KERNEL,
                NORM,
                beta,
            )

            if is_complex:
                x_range_cplx = x_range  # [width]
                y_range_cplx = y_range
                x_mask_cplx = x_mask
                y_mask_cplx = y_mask
                # Pytorch interleaved indexing
                # Only applies to last dimension
                z_range_real = 2 * z_range  # 2 is for real/complex, not dimension
                z_range_imag = 2 * z_range + 1
                z_range_cplx = tl.join(z_range_real, z_range_imag)  # [width, 2]
                z_mask_cplx = tl.join(z_mask, z_mask)
                if PAD_MODE == "zero":
                    x_mask_cplx &= (x_range_cplx >= 0) & (x_range_cplx < x_size)
                    y_mask_cplx &= (y_range_cplx >= 0) & (y_range_cplx < y_size)
                    z_mask_cplx &= (z_range_cplx >= 0) & (z_range_cplx < (2 * z_size))
                elif PAD_MODE == "circular":
                    x_range_cplx = mod_pos(x_range_cplx, x_size)
                    y_range_cplx = mod_pos(y_range_cplx, y_size)
                    z_range_cplx = mod_pos(z_range_cplx, 2 * z_size)

                grid_range_cplx = (
                    x_range_cplx[:, None, None, None] * y_size
                    + y_range_cplx[None, :, None, None]
                ) * z_size * 2 + z_range_cplx[None, None, :, :]
                grid_mask_cplx = (
                    x_mask_cplx[:, None, None, None] & y_mask_cplx[None, :, None, None]
                ) & z_mask_cplx[None, None, :, :]

                # Load split
                pt_real = tl.load(in_ptr + 2 * (in_batch_offset + p))
                pt_imag = tl.load(in_ptr + 2 * (in_batch_offset + p) + 1)
                out_real = weights * pt_real
                out_imag = weights * pt_imag
                out_cplx = tl.join(out_real, out_imag)

                # Accmulate
                tl.atomic_add(
                    out_ptr + out_batch_offset + grid_range_cplx,
                    out_cplx,
                    grid_mask_cplx,
                )

            else:
                # Normal indexing
                if PAD_MODE == "zero":
                    x_mask &= (x_range >= 0) & (x_range < x_size)
                    y_mask &= (y_range >= 0) & (y_range < y_size)
                    z_mask &= (z_range >= 0) & (z_range < z_size)
                elif PAD_MODE == "circular":
                    x_range = mod_pos(x_range, x_size)
                    y_range = mod_pos(y_range, y_size)
                    z_range = mod_pos(z_range, z_size)

                grid_range = (
                    x_range[:, None, None] * y_size + y_range[None, :, None]
                ) * z_size + z_range[None, None, :]
                grid_mask = (
                    x_mask[:, None, None]
                    & y_mask[None, :, None]
                    & z_mask[None, None, :]
                )

                # Load
                pt_val = tl.load(in_ptr + in_batch_offset + p)
                out = weights * pt_val

                # Accumulate
                tl.atomic_add(out_ptr + out_batch_offset + grid_range, out, grid_mask)


if TRITON_ENABLED:
    GRID = {1: _grid1d, 2: _grid2d, 3: _grid3d}
else:
    GRID = {}


def prep_grid_shapes(vals, locs, grid_size, width):
    ndim = len(grid_size)
    if locs.shape[-1] != ndim:
        raise ValueError(
            f"Expected last dim of locs: to equal ndim but got locs.shape[-1] {locs.shape[-1]} != ndim {ndim}"
        )
    # Handle locs shapes
    locs_batch_shape = locs.shape[:-1]
    npts = prod(locs_batch_shape)

    # Ensure locs are in [0, grid_size-1] in each dimension
    locs = torch.remainder(locs, torch.tensor(grid_size, device=locs.device))

    # Flatten input vals
    batch_shape = vals.shape[: -len(locs_batch_shape)]
    if len(locs_batch_shape) < len(vals.shape):
        vals_flat = vals.flatten(0, len(vals.shape) - len(locs_batch_shape) - 1)
    else:
        vals_flat = vals[None]
    nbatch = vals_flat.shape[0]

    # Complex input
    is_complex = torch.is_complex(vals)

    # Ensure kernel width is a tuple
    if isinstance(width, float):
        width = (width,) * ndim
    elif isinstance(width, tuple) and len(width) != ndim:
        raise ValueError(
            f"If width specified as tuple it must be same length as grid size but got len(width) = {len(width)} and len(grid_size) = {ndim}"
        )

    return (
        vals_flat,
        locs,
        {
            "ndim": ndim,
            "grid_size": grid_size,
            "width": width,
            "nbatch": nbatch,
            "batch_shape": batch_shape,
            "is_complex": is_complex,
            "locs_batch_shape": locs_batch_shape,
            "npts": npts,
        },
    )


def grid_torch(
    vals: Inexact[Tensor, "B ..."],
    locs: Float[Tensor, "... D"],
    grid_size: tuple[int, ...],
    width: tuple[float, ...],
    kernel: str = "kaiser_bessel",
    norm: str = "1",
    pad_mode: Literal["zero", "circular"] = "circular",
    batch_size: int = 2**20,
    **kernel_params,
):
    """Torch fallback

    Eventually, may want to use triton's CPU backend

    pad_mode : 'zero' or 'circular'
        Type of edge behavior to use
    batch_size : int
        number of points to compute over at once
    """

    kernel_fn = get_kernel_fn(kernel, kernel_params)

    # Define helper vars
    ndim = locs.shape[-1]
    device = vals.device
    nbatch = vals.shape[0]
    diff = torch.meshgrid(*(torch.arange(w + 1) for w in width), indexing="ij")
    # patch_size = diff[0].shape
    grid_size = torch.tensor(grid_size, device=vals.device)

    # [prod(patch_size), ndim]
    diff = torch.stack(diff, axis=-1).reshape(-1, ndim).to(device)
    radius = torch.tensor(width, device=device) / 2
    locs_lower = torch.ceil(locs - radius).to(torch.int64)

    # For loop for memory purposes
    out = torch.zeros(nbatch, *tuple(grid_size), dtype=vals.dtype, device=device)
    vals = vals.reshape(nbatch, -1)  # [nbatch, npts]
    npts = vals.shape[1]
    locs = locs.reshape(-1, ndim)  # [npts, ndim]
    locs_lower = locs_lower.reshape(-1, ndim)  # [npts, ndim]
    # Prepare batch indices (memory efficient because of expand())
    batch_indices = torch.arange(nbatch).view(-1, 1, 1)  # [B npts ngrid]
    for p0, p1 in batch_iterator(npts, batch_size):
        # [npts, npatch, ndim]
        grid_locs = locs_lower[p0:p1, None] + diff
        # Normalized delta(locations)
        weights, grid_locs, mask = weights_torch(
            locs[p0:p1],
            grid_locs,
            radius,
            norm,
            kernel_fn,
            grid_size,
            pad_mode,
        )
        val = vals[:, p0:p1, None]  # [nbatch, npts, 1]
        patches = weights * mask * val

        # Expand patch indices
        npts = p1 - p0
        ngrid = patches.shape[-1]
        batch_slc = batch_indices.expand(-1, npts, ngrid)

        # Perform indexing
        grid_locs_slc = (batch_slc, *(grid_locs[..., i] for i in range(ndim)))
        out.index_put_(grid_locs_slc, patches, accumulate=True)
    return out



import sys
import traceback
from collections.abc import Mapping
from typing import Optional

import torch
import torch.nn as nn

from torchlinops.utils import INDENT

from .nameddim import NS, isequal, NamedDimension as ND
from .namedlinop import NamedLinop


class Chain(NamedLinop):
    """A sequence or composition of linops"""

    def __init__(self, *linops, name: Optional[str] = None):
        """
        Parameters
        ----------
        *linops : list
            Linops in order of execution
            i.e. if `linops = [A, B, C]`, then mathematically, the linop in question is `CBA`

        """
        super().__init__(NS(linops[0].ishape, linops[-1].oshape), name=name)
        self.linops = nn.ModuleList(list(linops))
        self._check_inputs_outputs()

    def _check_inputs_outputs(self):
        curr_shape = self.ishape
        for i, linop in enumerate(self.linops):
            if not isequal(linop.ishape, curr_shape):
                raise ValueError(
                    f"Mismatched shape: expected {linop.ishape}, got {curr_shape} at input to {linop}. Full stack: {self}, index {i}"
                )
            curr_shape = linop.oshape

    def forward(self, x):
        for linop in self.linops:
            x = linop(x)
        return x

    @staticmethod
    def fn(chain, x: torch.Tensor, /, data_list):
        assert len(chain.linops) == len(data_list), (
            f"Length {len(data_list)} data_list does not match length {len(chain.linops)} chain linop"
        )
        for linop, data in zip(chain.linops, data_list):
            x = linop.fn(x, *data)
        return x

    @staticmethod
    def adj_fn(chain, x: torch.Tensor, /, data_list):
        assert len(chain.linops) == len(data_list), (
            f"Length {len(data_list)} data_list does not match length {len(chain.linops)} chain adjoint linop"
        )
        for linop, data in zip(reversed(chain.linops), reversed(data_list)):
            x = linop.adj_fn(x, data)
        return x

    @staticmethod
    def normal_fn(chain, x: torch.Tensor, /, data_list):
        # fn does the reversing so it's unnecessary to do it here
        # If the normal hasn't been explicitly formed with`.N`, do things the naive way
        return chain.adj_fn(chain.fn(x, data_list), data_list)

    def split_forward(self, ibatches, obatches):
        """ibatches, obatches specified according to the shape of the
        forward op
        """
        linops = [
            linop.split_forward(ibatch, obatch)
            for linop, ibatch, obatch in zip(self.linops, ibatches, obatches)
        ]
        return type(self)(*linops, name=self._name)

    def size(self, dim):
        out = None
        for linop in self.linops:
            tmp = linop.size(dim)
            if tmp is not None:
                if out is None:
                    out = tmp
                elif out != tmp:
                    raise ValueError(
                        f"Conflicting linop sizes found: {out} and {tmp} for dim {dim} in linop {linop} out of all linops {self.linops}"
                    )
        return out

    def size_fn(self, dim, data):
        for linop, data in zip(self.linops, data):
            out = linop.size_fn(dim, data)
            if out is not None:
                return out
        return None

    @property
    def dims(self):
        """Get the dims that appear anywhere in this linop chain."""
        return set().union(*[linop.dims for linop in self.linops])

    def adjoint(self):
        linops = list(linop.adjoint() for linop in reversed(self.linops))
        return type(self)(*linops, name=self._name)

    def normal(self, inner=None):
        for linop in reversed(self.linops):
            inner = linop.normal(inner)
        return inner

    @staticmethod
    def split(chain, tile: Mapping[ND | str, slice]):
        """Split a linop into sub-linops.

        Parameters
        ----------
        chain : Chain
            The chain linop to split.
        tile : Mapping[ND | str, slice]
            Dictionary specifying how to slice the linop dimensions
        """
        ibatches = [
            [tile.get(dim, slice(None)) for dim in linop.ishape]
            for linop in chain.linops
        ]
        obatches = [
            [tile.get(dim, slice(None)) for dim in linop.oshape]
            for linop in chain.linops
        ]
        return chain.split_forward(ibatches, obatches)

    @staticmethod
    def adj_split(chain, tile: Mapping[ND | str, slice]):
        """Split an adjoint linop into sub-linops.

        Parameters
        ----------
        chain : Chain
            The chain linop to split.
        tile : Mapping[ND | str, slice]
            Dictionary specifying how to slice the linop dimensions
        """
        ibatches = [
            [tile.get(dim, slice(None)) for dim in linop.ishape]
            for linop in chain.linops
        ]
        obatches = [
            [tile.get(dim, slice(None)) for dim in linop.oshape]
            for linop in chain.linops
        ]
        return chain.H.split_forward(obatches, ibatches).H

    # DEPRECATED/TODO: Remove
    # @staticmethod
    # def split_fn(chain, *iobatchesdata):
    #     """Return split versions of the data that can be passed
    #     into fn and adj_fn to produce split versions
    #     """
    #     ibatches = iobatchesdata[: len(iobatchesdata) // 3]
    #     obatches = iobatchesdata[len(iobatchesdata) // 3 : len(iobatchesdata) * 2 // 3]
    #     data = iobatchesdata[len(iobatchesdata) * 2 // 3 :]
    #     return chain.split_forward_fn(ibatches, obatches, data)

    # @staticmethod
    # def adj_split_fn(chain, *iobatchesdata):
    #     ibatches = iobatchesdata[: len(iobatchesdata) // 3]
    #     obatches = iobatchesdata[len(iobatchesdata) // 3 : len(iobatchesdata) * 2 // 3]
    #     data = iobatchesdata[len(iobatchesdata) * 2 // 3]
    #     return chain.split_forward_fn(obatches, ibatches, data)

    @property
    def shape(self):
        return NS(self.linops[0].ishape, self.linops[-1].oshape)

    @shape.setter
    def shape(self, val):
        self.ishape = val.ishape
        self.oshape = val.oshape

    @property
    def ishape(self):
        return self.linops[0].ishape

    @ishape.setter
    def ishape(self, val):
        self.linops[0].ishape = val

    @property
    def oshape(self):
        return self.linops[-1].oshape

    @oshape.setter
    def oshape(self, val):
        self.linops[-1].oshape = val

    def flatten(self):
        return list(self.linops)

    def __getitem__(self, idx):
        linops = self.linops[idx]
        if isinstance(linops, NamedLinop):
            return linops
        return Chain(*linops, name=self._name)

    def __len__(self):
        return len(self.linops)

    def __repr__(self):
        output = ""
        output += INDENT.indent(self.repr_name + "(\n")
        with INDENT:
            for linop in self.linops:
                output += repr(linop) + "\n"
        output += INDENT.indent(")")
        return output



"""Recursive mapping on data with function"""

from typing import Callable, Mapping, Optional

import torch
import numpy as np

__all__ = [
    "apply_struct",
    "numpy2torch",
    "print_shapes",
]


def apply_struct(struct, fn: Callable, condition: Callable):
    if isinstance(struct, Mapping):
        kv_pairs = struct.items()
    elif isinstance(struct, list):
        kv_pairs = enumerate(struct)
    elif condition(struct):
        return fn(struct)
    else:
        return struct
        # raise NotImplementedError(f'Struct should be a dict or a list (got {type(struct)})')
    for k, v in kv_pairs:
        struct[k] = apply_struct(v, fn, condition)
    return struct


def numpy2torch(data, device: Optional[torch.device] = "cpu"):
    return apply_struct(
        data,
        lambda x: torch.from_numpy(x).to(device),
        lambda x: isinstance(x, np.ndarray),
    )


def torch2numpy(data):
    return apply_struct(
        data,
        lambda x: x.detach().cpu().numpy(),
        lambda x: isinstance(x, torch.Tensor),
    )


def print_shapes(data):
    for name, obj in data.items():
        print(f"{name}: {obj.shape}")



from typing import Optional
from jaxtyping import Shaped, Bool, Float
from torch import Tensor

from itertools import product

import torch

try:
    import triton
    import triton.language as tl
    from .casting import scalar_cast as cast

    TRITON_ENABLED = True
except ImportError:
    from torchlinops.utils import fake_triton as triton, fake_tl as tl

    TRITON_ENABLED = False

from .nblocks import get_nblocks

__all__ = ["unfold"]

# Maximum size of the cuda grid per dimension
# 3 dimensions maximum
MAX_GRID_PER_DIM = 1024

MAX_TENSOR_POW_OF_2 = 20


def unfold(
    x: Shaped[Tensor, "..."],
    block_size: tuple,
    stride: Optional[tuple] = None,
    mask: Optional[Bool[Tensor, "..."]] = None,
    output: Optional[Tensor] = None,
) -> Tensor:
    """Wrapper that dispatches complex and real tensors
    Also precomputes some shapes

    Parameters
    ----------
    x : Tensor
        Shape [B..., *im_size]

    Returns
    -------
    Tensor: Shape [B..., *blocks, *block_size]
        If mask is not None, block_size will be an int equal to the number of True elements in the mask
        Otherwise it will be the full block shape.


    """
    x_flat, shapes, is_complex = _prep_unfold(x, block_size, stride, mask)
    if is_complex:
        x_flat = torch.view_as_real(x_flat)
        x_flat = torch.flatten(x_flat, -2, -1)  # Flatten real/imag into last dim
    y_flat = _unfold(x_flat, output=output, **shapes)
    y = y_flat.reshape(
        *shapes["batch_shape"],
        *shapes["nblocks"],
        *shapes["block_size"],
    )
    if is_complex:
        y = y.reshape(*y.shape[:-1], y.shape[-1] // 2, 2)
        y = torch.view_as_complex(y)
    if mask is not None:
        y = y[..., mask]
    return y


def _unfold(
    x: Shaped[Tensor, "B ..."],
    block_size: tuple[int, ...],
    stride: tuple[int, ...],
    ndim: int,
    im_size: tuple[int, ...],
    nblocks: tuple[int, ...],
    nbatch: int,
    output: Optional[Tensor] = None,
    **kwargs,
) -> Shaped[Tensor, "B ..."]:
    """Implementation of unfold"""
    # Check dtype if output buffer is provided
    if output is not None:
        if not output.dtype == x.dtype:
            raise ValueError(
                f"Output and input dtypes must match but got output {output.dtype} != input {x.dtype}"
            )
    if tuple(x.shape[-ndim:]) != tuple(im_size):
        raise RuntimeError(
            f"Unfold expected input with full size {im_size} but got {x.shape}"
        )
    if x.is_cuda and ndim in UNFOLD.keys():
        x = x.contiguous()  # Ensure contiguity
        with torch.cuda.device(x.device):
            if output is None:
                # Allocate output
                y = torch.zeros(
                    nbatch,
                    *nblocks,
                    *block_size,
                    device=x.device,
                    dtype=x.dtype,
                )
            else:
                # Use existing buffer
                y = output.reshape(nbatch, *nblocks, *block_size).zero_()
            grid = _get_grid(ndim, nbatch, nblocks)
            BLOCK_SIZE = tuple(
                min(
                    triton.next_power_of_2(blk_size),
                    2 ** (MAX_TENSOR_POW_OF_2 // ndim),
                )
                for blk_size in block_size
            )
            UNFOLD[ndim][grid](
                x,
                y,
                nbatch,
                *nblocks,
                *block_size,
                *im_size,
                *stride,
                *BLOCK_SIZE,
            )
    else:
        y = _unfold_torch(
            x, block_size, stride, ndim, im_size, nblocks, nbatch, out=output
        )
    return y


def _get_grid(ndim: int, nbatch, nblocks: tuple[int, ...]):
    if ndim == 1:
        grid = lambda meta: (  # noqa: E731
            nbatch * triton.cdiv(nblocks[0], meta["x_blocks_per_grid"]),
        )
    elif ndim == 2:
        grid = lambda meta: (  # noqa: E731
            nbatch * triton.cdiv(nblocks[0], meta["x_blocks_per_grid"]),
            triton.cdiv(nblocks[1], meta["y_blocks_per_grid"]),
        )
    elif ndim == 3:
        grid = lambda meta: (  # noqa: E731
            nbatch * triton.cdiv(nblocks[0], meta["x_blocks_per_grid"]),
            triton.cdiv(nblocks[1], meta["y_blocks_per_grid"]),
            triton.cdiv(nblocks[2], meta["z_blocks_per_grid"]),
        )
    else:
        raise ValueError(f"Invalid ndim = {ndim}")
    return grid


@triton.heuristics(
    values={
        "x_blocks_per_grid": lambda args: max(
            1, triton.cdiv(args["x_nblocks"], MAX_GRID_PER_DIM)
        ),
        "x_BLOCKS_per_block": lambda args: max(
            1, triton.cdiv(args["x_block_dim"], 2**MAX_TENSOR_POW_OF_2)
        ),
    },
)
@triton.jit
def _unfold1d(
    in_ptr,
    out_ptr,
    # Number of batches
    nbatch: int,
    # Number of blocks
    x_nblocks: int,
    # Size of each block
    x_block_dim: int,
    # Size of the input data
    x_size: int,
    # Stride of the blocks
    x_stride: int,
    # Size of the triton block (power of 2)
    X_BLOCK_SIZE: tl.constexpr,
    # Number of blocks per grid pid
    x_blocks_per_grid: int,
    # Number of triton blocks per block
    x_BLOCKS_per_block: int,
):
    """
    Note: Cannot use make_block_ptr for out_ptr because the output block
    might require masking.
    """
    pid_0 = tl.program_id(0)
    # Batch index, Block index
    NBx = pid_0 * x_blocks_per_grid
    N, Bx = NBx // x_nblocks, NBx % x_nblocks

    # Convert types
    x_nblocks = cast(x_nblocks, tl.uint64)
    x_size = cast(x_size, tl.uint64)
    x_block_dim = cast(x_block_dim, tl.uint64)

    in_size = x_size
    nblocks = x_nblocks
    block_dim = x_block_dim

    in_blk_ptr = tl.make_block_ptr(
        in_ptr,
        shape=(nbatch, x_size),
        strides=(in_size, 1),
        offsets=(N, Bx * x_stride),
        block_shape=(1, X_BLOCK_SIZE),
        order=(0, 1),
    )
    x_base_range = tl.arange(0, X_BLOCK_SIZE)
    for i in range(x_blocks_per_grid):
        if Bx + i < x_nblocks:
            for u in range(x_BLOCKS_per_block):
                x_range = x_base_range + u * X_BLOCK_SIZE
                x_mask = x_range < x_block_dim
                blk_range = x_range
                blk_mask = x_mask

                out_range = blk_range[None]
                out_mask = blk_mask[None]
                blk = load_subblock1d(in_blk_ptr, u, X_BLOCK_SIZE)
                # Save block to output
                out_offset = N * nblocks * block_dim + (Bx + i) * block_dim
                tl.store(out_ptr + out_offset + out_range, blk, out_mask)
        in_blk_ptr = tl.advance(in_blk_ptr, (0, x_stride))


@triton.jit
def load_subblock1d(in_blk_ptr, x_idx: int, X_BLOCK_SIZE: tl.constexpr):
    return tl.load(
        in_blk_ptr.advance((0, x_idx * X_BLOCK_SIZE)),
        boundary_check=(1,),
        padding_option="zero",
    )


@triton.heuristics(
    values={
        "x_blocks_per_grid": lambda args: max(
            1, triton.cdiv(args["x_nblocks"], MAX_GRID_PER_DIM)
        ),
        "y_blocks_per_grid": lambda args: max(
            1, triton.cdiv(args["y_nblocks"], MAX_GRID_PER_DIM)
        ),
        "x_BLOCKS_per_block": lambda args: max(
            1, triton.cdiv(args["x_block_dim"], 2 ** (MAX_TENSOR_POW_OF_2 // 2))
        ),
        "y_BLOCKS_per_block": lambda args: max(
            1, triton.cdiv(args["y_block_dim"], 2 ** (MAX_TENSOR_POW_OF_2 // 2))
        ),
    },
)
@triton.jit
def _unfold2d(
    in_ptr,
    out_ptr,
    # Number of batches
    nbatch: int,
    # Number of blocks
    x_nblocks: int,
    y_nblocks: int,
    # Size of each block
    x_block_dim: int,
    y_block_dim: int,
    # Size of the input data
    x_size: int,
    y_size: int,
    # Stride of the blocks
    x_stride: int,
    y_stride: int,
    # Size of the triton block (power of 2)
    X_BLOCK_SIZE: tl.constexpr,
    Y_BLOCK_SIZE: tl.constexpr,
    # Number of blocks per grid pid
    x_blocks_per_grid: int,
    y_blocks_per_grid: int,
    # Number of triton blocks per block
    x_BLOCKS_per_block: int,
    y_BLOCKS_per_block: int,
):
    pid_0 = tl.program_id(0)
    pid_1 = tl.program_id(1)
    # Batch index, Block index
    NBx = pid_0 * x_blocks_per_grid
    N, Bx = NBx // x_nblocks, NBx % x_nblocks
    By = pid_1 * y_blocks_per_grid

    # Convert types
    x_nblocks = cast(x_nblocks, tl.uint64)
    y_nblocks = cast(y_nblocks, tl.uint64)
    x_size = cast(x_size, tl.uint64)
    y_size = cast(y_size, tl.uint64)
    x_block_dim = cast(x_block_dim, tl.uint64)
    y_block_dim = cast(y_block_dim, tl.uint64)

    # global sizes
    in_size = x_size * y_size
    nblocks = x_nblocks * y_nblocks
    block_dim = x_block_dim * y_block_dim

    in_blk_ptr = tl.make_block_ptr(
        in_ptr,
        shape=(nbatch, x_size, y_size),
        strides=(in_size, y_size, 1),
        offsets=(N, Bx * x_stride, By * y_stride),
        block_shape=(1, X_BLOCK_SIZE, Y_BLOCK_SIZE),
        order=(0, 1, 2),
    )
    x_base_range = tl.arange(0, X_BLOCK_SIZE)
    y_base_range = tl.arange(0, Y_BLOCK_SIZE)

    for i in range(x_blocks_per_grid):
        if Bx + i < x_nblocks:
            x_blk_offset = (Bx + i) * y_nblocks * block_dim
            in_blk_ptr_x = in_blk_ptr
            for j in range(y_blocks_per_grid):
                if By + j < y_nblocks:
                    y_blk_offset = (By + j) * block_dim
                    # Loop over blocks within block
                    for u in range(x_BLOCKS_per_block):
                        for v in range(y_BLOCKS_per_block):
                            x_range = x_base_range + u * X_BLOCK_SIZE
                            y_range = y_base_range + v * Y_BLOCK_SIZE
                            x_mask = x_range < x_block_dim
                            y_mask = y_range < y_block_dim
                            blk_range = (
                                x_range[:, None] * y_block_dim + y_range[None, :]
                            )
                            blk_mask = x_mask[:, None] & y_mask[None, :]
                            # out_offset = N * x_nblocks * x_block_dim + Bx * x_block_dim
                            # add batch dim
                            out_range = blk_range[None, :, :]
                            out_mask = blk_mask[None, :, :]
                            # blk = tl.load(in_blk_ptr)
                            blk = load_subblock2d(
                                in_blk_ptr, u, v, X_BLOCK_SIZE, Y_BLOCK_SIZE
                            )
                            out_offset = (
                                N * nblocks * block_dim + x_blk_offset + y_blk_offset
                            )
                            tl.store(out_ptr + out_offset + out_range, blk, out_mask)
                in_blk_ptr = tl.advance(in_blk_ptr, (0, 0, y_stride))
            in_blk_ptr = in_blk_ptr_x
        in_blk_ptr = tl.advance(in_blk_ptr, (0, x_stride, 0))


@triton.jit
def load_subblock2d(in_blk_ptr, x_idx, y_idx, X_BLOCK_SIZE, Y_BLOCK_SIZE):
    return tl.load(
        in_blk_ptr.advance((0, x_idx * X_BLOCK_SIZE, y_idx * Y_BLOCK_SIZE)),
        boundary_check=(1, 2),
        padding_option="zero",
    )


@triton.heuristics(
    values={
        "x_blocks_per_grid": lambda args: max(
            1, triton.cdiv(args["x_nblocks"], MAX_GRID_PER_DIM)
        ),
        "y_blocks_per_grid": lambda args: max(
            1, triton.cdiv(args["y_nblocks"], MAX_GRID_PER_DIM)
        ),
        "z_blocks_per_grid": lambda args: max(
            1, triton.cdiv(args["z_nblocks"], MAX_GRID_PER_DIM)
        ),
        "x_BLOCKS_per_block": lambda args: max(
            1, triton.cdiv(args["x_block_dim"], 2 ** (MAX_TENSOR_POW_OF_2 // 3))
        ),
        "y_BLOCKS_per_block": lambda args: max(
            1, triton.cdiv(args["y_block_dim"], 2 ** (MAX_TENSOR_POW_OF_2 // 3))
        ),
        "z_BLOCKS_per_block": lambda args: max(
            1, triton.cdiv(args["z_block_dim"], 2 ** (MAX_TENSOR_POW_OF_2 // 3))
        ),
    },
)
@triton.jit
def _unfold3d(
    in_ptr,
    out_ptr,
    # Number of batches
    nbatch: int,
    # Number of blocks
    x_nblocks: int,
    y_nblocks: int,
    z_nblocks: int,
    # Size of each block
    x_block_dim: int,
    y_block_dim: int,
    z_block_dim: int,
    # Size of the input data
    x_size: int,
    y_size: int,
    z_size: int,
    # Stride of the blocks
    x_stride: int,
    y_stride: int,
    z_stride: int,
    # Size of the triton block (power of 2)
    X_BLOCK_SIZE: tl.constexpr,
    Y_BLOCK_SIZE: tl.constexpr,
    Z_BLOCK_SIZE: tl.constexpr,
    # Number of blocks per grid pid
    x_blocks_per_grid: int,
    y_blocks_per_grid: int,
    z_blocks_per_grid: int,
    # Number of triton blocks per block
    x_BLOCKS_per_block: int,
    y_BLOCKS_per_block: int,
    z_BLOCKS_per_block: int,
):
    """"""
    pid_0 = tl.program_id(0)
    pid_1 = tl.program_id(1)
    pid_2 = tl.program_id(2)
    # Batch index, Block index
    NBx = pid_0 * x_blocks_per_grid
    N, Bx = NBx // x_nblocks, NBx % x_nblocks
    By = pid_1 * y_blocks_per_grid
    Bz = pid_2 * z_blocks_per_grid

    # Convert types
    x_nblocks = cast(x_nblocks, tl.uint64)
    y_nblocks = cast(y_nblocks, tl.uint64)
    z_nblocks = cast(z_nblocks, tl.uint64)
    x_size = cast(x_size, tl.uint64)
    y_size = cast(y_size, tl.uint64)
    z_size = cast(z_size, tl.uint64)
    x_block_dim = cast(x_block_dim, tl.uint64)
    y_block_dim = cast(y_block_dim, tl.uint64)
    z_block_dim = cast(z_block_dim, tl.uint64)

    # global sizes
    in_size = x_size * y_size * z_size
    nblocks = x_nblocks * y_nblocks * z_nblocks
    block_dim = x_block_dim * y_block_dim * z_block_dim

    in_blk_ptr = tl.make_block_ptr(
        in_ptr,
        shape=(nbatch, x_size, y_size, z_size),
        strides=(in_size, y_size * z_size, z_size, 1),
        offsets=(N, Bx * x_stride, By * y_stride, Bz * z_stride),
        block_shape=(1, X_BLOCK_SIZE, Y_BLOCK_SIZE, Z_BLOCK_SIZE),
        order=(0, 1, 2, 3),
    )
    x_base_range = tl.arange(0, X_BLOCK_SIZE)
    y_base_range = tl.arange(0, Y_BLOCK_SIZE)
    z_base_range = tl.arange(0, Z_BLOCK_SIZE)

    for i in range(x_blocks_per_grid):
        if Bx + i < x_nblocks:
            x_blk_offset = (Bx + i) * y_nblocks * z_nblocks * block_dim
            in_blk_ptr_x = in_blk_ptr
            for j in range(y_blocks_per_grid):
                if By + j < y_nblocks:
                    y_blk_offset = (By + j) * z_nblocks * block_dim
                    in_blk_ptr_y = in_blk_ptr
                    for k in range(z_blocks_per_grid):
                        if Bz + k < z_nblocks:
                            z_blk_offset = (Bz + k) * block_dim
                            # Loop over subblocks
                            for u in range(x_BLOCKS_per_block):
                                for v in range(y_BLOCKS_per_block):
                                    for w in range(z_BLOCKS_per_block):
                                        x_range = x_base_range + u * X_BLOCK_SIZE
                                        y_range = y_base_range + v * Y_BLOCK_SIZE
                                        z_range = z_base_range + w * Z_BLOCK_SIZE
                                        x_mask = x_range < x_block_dim
                                        y_mask = y_range < y_block_dim
                                        z_mask = z_range < z_block_dim
                                        blk_range = (
                                            x_range[:, None, None] * y_block_dim
                                            + y_range[None, :, None]
                                        ) * z_block_dim + z_range[None, None, :]
                                        blk_mask = x_mask[:, None, None] & (
                                            y_mask[None, :, None]
                                            & z_mask[None, None, :]
                                        )

                                        out_range = blk_range[None]
                                        out_mask = blk_mask[None]
                                        blk = load_subblock3d(
                                            in_blk_ptr,
                                            u,
                                            v,
                                            w,
                                            X_BLOCK_SIZE,
                                            Y_BLOCK_SIZE,
                                            Z_BLOCK_SIZE,
                                        )
                                        out_offset = (
                                            N * nblocks * block_dim
                                            + x_blk_offset
                                            + y_blk_offset
                                            + z_blk_offset
                                        )
                                        tl.store(
                                            out_ptr + out_offset + out_range,
                                            blk,
                                            out_mask,
                                        )
                        in_blk_ptr = tl.advance(in_blk_ptr, (0, 0, 0, z_stride))
                    in_blk_ptr = in_blk_ptr_y
                in_blk_ptr = tl.advance(in_blk_ptr, (0, 0, y_stride, 0))
            in_blk_ptr = in_blk_ptr_x
        in_blk_ptr = tl.advance(in_blk_ptr, (0, x_stride, 0, 0))


@triton.jit
def load_subblock3d(
    in_blk_ptr, x_idx, y_idx, z_idx, X_BLOCK_SIZE, Y_BLOCK_SIZE, Z_BLOCK_SIZE
):
    return tl.load(
        in_blk_ptr.advance(
            (0, x_idx * X_BLOCK_SIZE, y_idx * Y_BLOCK_SIZE, z_idx * Z_BLOCK_SIZE)
        ),
        boundary_check=(1, 2, 3),
        padding_option="zero",
    )


if TRITON_ENABLED:
    UNFOLD = {1: _unfold1d, 2: _unfold2d, 3: _unfold3d}
else:
    UNFOLD = {}


def _unfold_torch(
    x: Float[Tensor, "B ..."],
    block_size: tuple[int, ...],
    stride: tuple[int, ...],
    ndim: int,
    im_size: tuple[int, ...],
    nblocks: tuple[int, ...],
    nbatch: int,
    out: Optional[Tensor] = None,
) -> Float[Tensor, "B I ..."]:
    """Fallback option

    Note: Compile takes forever
    """
    if out is None:
        out = torch.zeros(
            (nbatch, *nblocks, *block_size), device=x.device, dtype=x.dtype
        )
    else:
        out = out.reshape(nbatch, *nblocks, *block_size).zero_()
    # Python implementation
    for batch in range(nbatch):
        for blk in product(*(range(nblk) for nblk in nblocks)):
            blk_slc = tuple(
                slice(iblk * st, iblk * st + blk_sz)
                for iblk, st, blk_sz in zip(blk, stride, block_size)
            )
            out_idx = (batch, *blk)
            in_idx = (batch, *blk_slc)
            out[out_idx] = x[in_idx]
    return out


def _prep_unfold(
    x,
    block_size: tuple,
    stride: Optional[tuple] = None,
    mask: Optional[Bool[Tensor, "..."]] = None,
):
    is_complex = torch.is_complex(x)
    # Infer some shapes
    ndim = len(block_size)
    im_size = x.shape[-ndim:]
    stride = stride if stride is not None else (1,) * ndim
    nblocks = get_nblocks(im_size, block_size, stride)
    if any(b == 0 for b in nblocks):
        raise ValueError(
            f"Found 0 in nblocks: {nblocks} with im_size {im_size}, block_size {block_size} and stride {stride} - make sure there is at least one block in each direction."
        )

    if is_complex:
        im_size = list(im_size)
        im_size[-1] *= 2
        block_size = list(block_size)
        block_size[-1] *= 2
        stride = list(stride)
        stride[-1] *= 2

    # Add or infer batch dim
    batch_shape = x.shape[:-ndim]
    if ndim < len(x.shape):
        x_flat = x.flatten(0, len(x.shape) - ndim - 1)
    else:
        x_flat = x[None]
    nbatch = x_flat.shape[0]

    # Handle mask
    if mask is not None:
        mask = mask.to(x.device)

    return (
        x_flat,
        {
            "ndim": ndim,
            "im_size": im_size,
            "stride": stride,
            "nblocks": nblocks,
            "nbatch": nbatch,
            "batch_shape": batch_shape,
            "mask": mask,
            "block_size": block_size,
        },
        is_complex,
    )



from jaxtyping import Float, Integer, Bool
from torch import Tensor
from typing import Literal
from collections.abc import Callable

from functools import partial

import torch

try:
    import triton
    import triton.language as tl

    TRITON_ENABLED = True
except ImportError:
    from torchlinops.utils import fake_triton as triton, fake_tl as tl

    TRITON_ENABLED = False

__all__ = [
    "kaiser_bessel",
    "kaiser_bessel_torch",
    "spline_torch",
    "spline",
]


@torch.compile
def kaiser_bessel_torch(x: Float[Tensor, "..."], beta: float):
    """Vectorized kaiser-bessel kernel implementation
    Parameters
    ----------
    x: Tensor, any shape
        Kernel acts on favlues in x in the range [-1, 1]
    beta: float
        Shape parameter for kaiser bessel kernel

    Explanation
    -----------
    The scary polynomial stuff is a polynomial approximation of the zeroth-order modified Bessel
    function of the first kind.
    Here, beta is standing in for a bunch of constant factors that multiply in front of it.

    References
    ----------
    Polynomial approximation of Bessel function from
    Abramowitz and Stegun, Handbook of Mathematical Functions, p.378
    https://personal.math.ubc.ca/~cbm/aands/page_378.htm
    """
    x = beta * torch.sqrt(1 - x**2)
    smallmask = x < 3.75
    t = x / 3.75
    # Optimized computation ordering
    t2 = t * t
    t4 = t2 * t2
    t6 = t4 * t2
    t8 = t4 * t4
    t10 = t6 * t4
    t12 = t6 * t6

    t3 = t2 * t
    t5 = t3 * t2
    t7 = t5 * t2

    small = (
        1
        + 3.5156229 * t2
        + 3.0899424 * t4
        + 1.2067492 * t6
        + 0.2659732 * t8
        + 0.0360768 * t10
        + 0.0045813 * t12
    )
    big = (
        torch.exp(x)
        / torch.sqrt(x)
        * (
            0.39894228
            + 0.01328592 / t
            + 0.00225319 / t2
            - 0.00157565 / t3
            + 0.00916281 / t4
            - 0.02057706 / t5
            + 0.02635537 / t6
            - 0.01647633 / t7
            + 0.00392377 / t8
        )
    )
    return torch.where(smallmask, small, big)


@torch.compile
def spline_torch(x):
    return torch.maximum(1.0 - torch.abs(x), torch.tensor(0.0))


def get_kernel_fn(kernel, kernel_params) -> Callable[[Tensor], Tensor]:
    if kernel == "spline":
        kernel_fn = spline_torch
    elif kernel == "kaiser_bessel":
        beta = kernel_params.get("beta", 1.0)
        kernel_fn = partial(kaiser_bessel_torch, beta=beta)
    else:
        raise ValueError(f"Unrecognized kernel {kernel}")
    return kernel_fn


def weights_torch(
    locs: Float[Tensor, "P G D"],
    grid_locs: Float[Tensor, "P G D"],
    radius: Float[Tensor, "D"],  # noqa: F821
    norm: Literal["1", "2"],
    kernel_fn: Callable,
    grid_size: Integer[Tensor, "D"],  # noqa: F821
    padding: Literal["zero", "circular"],
) -> tuple[Float[Tensor, "P G"], Integer[Tensor, "P G D"], Bool[Tensor, "P G"]]:
    """
    P: Number of pts
    G: Patch size (flattened)
    D: Dim (e.g. 3D)

    norm : str, 1 or 2
        Type of norm to use
    """
    # Normalized delta(locations)
    dlocs = (locs[:, None] - grid_locs) / radius

    # Weights shape: [npts, npatch]
    # Mask shape: [npts, npatch]
    if norm == "2":
        dist = torch.linalg.vector_norm(dlocs, dim=-1)
        weights = kernel_fn(dist)
        mask = dist <= radius
    elif norm == "1":
        weights = kernel_fn(dlocs).prod(dim=-1)
        mask = (torch.abs(dlocs) <= radius).all(dim=-1)
    else:
        raise ValueError(f"Unexpected norm: {norm}.")
    weights[~weights.isfinite()] = 0.0

    # Edge behavior
    if padding == "zero":
        mask &= (grid_locs >= 0).all(dim=-1) & (grid_locs < grid_size).all(dim=-1)
    # This does circular padding! unless the mask above is active
    grid_locs = torch.remainder(grid_locs, grid_size)
    grid_locs = grid_locs.to(torch.int64)
    return weights, grid_locs, mask


# Euclidean norm
@triton.jit
def norm1d(vx):
    return tl.abs(vx)


@triton.jit
def norm2d(vx, vy):
    absvx = tl.abs(vx)
    absvy = tl.abs(vy)
    return tl.sqrt(absvx * absvx + absvy * absvy)


@triton.jit
def norm3d(vx, vy, vz):
    absvx = tl.abs(vx)
    absvy = tl.abs(vy)
    absvz = tl.abs(vz)
    return tl.sqrt(absvx * absvx + absvy * absvy + absvz * absvz)


@triton.jit
def kaiser_bessel(x, beta):
    """Vectorized kaiser-bessel kernel implementation
    Parameters
    ----------
    x: Tensor, any shape
        Kernel acts on favlues in x in the range [-1, 1]
    beta: float
        Shape parameter for kaiser bessel kernel
        1.0 is a good default
        Ultimately, should depend on NUFFT width/oversampling

    Explanation
    -----------
    The scary polynomial stuff is a polynomial approximation of the zeroth-order modified Bessel
    function of the first kind.
    Here, beta is standing in for a bunch of constant factors that multiply in front of it.
    If using a NUFFT, beta can be determined via the formula

    `beta = np.pi * (((width / oversamp) * (oversamp - 0.5)) ** 2 - 0.8) ** 0.5`

    References
    ----------
    Polynomial approximation of Bessel function from
    Abramowitz and Stegun, Handbook of Mathematical Functions, p.378
    https://personal.math.ubc.ca/~cbm/aands/page_378.htm
    """
    x = beta * tl.sqrt(1 - x * x)
    smallmask = x < 3.75
    t = x / 3.75
    # Optimized computation ordering
    t2 = t * t
    t4 = t2 * t2
    t6 = t4 * t2
    t8 = t4 * t4
    t10 = t6 * t4
    t12 = t6 * t6

    t3 = t2 * t
    t5 = t3 * t2
    t7 = t5 * t2

    small = (
        1
        + 3.5156229 * t2
        + 3.0899424 * t4
        + 1.2067492 * t6
        + 0.2659732 * t8
        + 0.0360768 * t10
        + 0.0045813 * t12
    )
    big = (
        tl.exp(x)
        / tl.sqrt(x)
        * (
            0.39894228
            + 0.01328592 / t
            + 0.00225319 / t2
            - 0.00157565 / t3
            + 0.00916281 / t4
            - 0.02057706 / t5
            + 0.02635537 / t6
            - 0.01647633 / t7
            + 0.00392377 / t8
        )
    )
    return tl.where(smallmask, small, big)


@triton.jit
def spline(x):
    return tl.maximum(1.0 - tl.abs(x), 0.0)


### Common functions for grid/ungrid
KernelTypeStr = Literal["kaiser_bessel", "spline"]


def _apply_default_kernel_params(kernel: KernelTypeStr, kernel_params: dict):
    if kernel == "kaiser_bessel":
        kernel_params["beta"] = kernel_params.get("beta", 1.0)
    elif kernel == "spline":
        pass
    else:
        raise ValueError(f"Unrecognized kernel type: {kernel}")
    return kernel_params


@triton.jit
def weights1d(
    x_target,
    x_kernel_width,  # width to load
    x_base_range,  # block array to use to load
    KERNEL: tl.constexpr,  # Kernel type
    beta,
):
    # Compute neighborhood of target point
    x_range = get_neighborhood(x_target, x_kernel_width, x_base_range)

    # Compute kernel weights
    dx, rx = x_range - x_target, x_kernel_width / 2.0
    x_mask = tl.abs(dx) <= rx
    d = norm1d(dx) / rx
    if KERNEL == "kaiser_bessel":
        weights = kaiser_bessel(d, beta)
    elif KERNEL == "spline":
        weights = spline(d)
    else:
        tl.device_assert(False, f"Invalid kernel type: {KERNEL}")
    # Mask off out-of-range points
    weights_mask = x_mask
    weights = tl.where(weights_mask, weights, 0.0)
    return weights, x_range, x_mask


@triton.jit
def weights2d(
    x_target,
    y_target,
    x_kernel_width,
    y_kernel_width,
    x_base_range,
    y_base_range,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    beta,
):
    # Compute neighborhood of target point
    x_range = get_neighborhood(x_target, x_kernel_width, x_base_range)
    y_range = get_neighborhood(y_target, y_kernel_width, y_base_range)

    # Compute kernel weights
    dx, rx = (x_range - x_target), (x_kernel_width / 2.0)
    x_mask = tl.abs(dx) <= rx
    dy, ry = (y_range - y_target), (y_kernel_width / 2.0)
    y_mask = tl.abs(dy) <= ry
    if NORM == "2":
        d = norm2d(dx / rx, dy / ry)
        if KERNEL == "kaiser_bessel":
            weights = kaiser_bessel(d, beta)
        elif KERNEL == "spline":
            weights = spline(d)
        else:
            tl.device_assert(False, f"Invalid kernel type: {KERNEL}")
    elif NORM == "1":
        if KERNEL == "kaiser_bessel":
            wx = kaiser_bessel(norm1d(dx) / rx, beta)
            wy = kaiser_bessel(norm1d(dy) / ry, beta)
            weights = wx[:, None] * wy[None, :]
        elif KERNEL == "spline":
            wx = spline(norm1d(dx) / rx)
            wy = spline(norm1d(dy) / ry)
            weights = wx[:, None] * wy[None, :]
        else:
            tl.device_assert(False, f"Invalid kernel type: {KERNEL}")
    # Mask off NaNs and out-of-range values
    weights_mask = x_mask[:, None] & y_mask[None, :]
    weights = tl.where(weights_mask, weights, 0.0)
    return weights, x_range, y_range, x_mask, y_mask


@triton.jit
def weights3d(
    x_target,
    y_target,
    z_target,
    x_kernel_width,
    y_kernel_width,
    z_kernel_width,
    x_base_range,
    y_base_range,
    z_base_range,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    beta,
):
    # Compute neighborhood of target point
    x_range = get_neighborhood(x_target, x_kernel_width, x_base_range)
    y_range = get_neighborhood(y_target, y_kernel_width, y_base_range)
    z_range = get_neighborhood(z_target, z_kernel_width, z_base_range)

    # Compute kernel weights
    dx, rx = (x_range - x_target), (x_kernel_width / 2.0)
    x_mask = tl.abs(dx) <= rx
    dy, ry = (y_range - y_target), (y_kernel_width / 2.0)
    y_mask = tl.abs(dy) <= ry
    dz, rz = (z_range - z_target), (z_kernel_width / 2.0)
    z_mask = tl.abs(dz) <= rz
    if NORM == "2":
        d = norm3d(dx / rx, dy / ry, dz / rz)
        if KERNEL == "kaiser_bessel":
            weights = kaiser_bessel(d, beta)
        elif KERNEL == "spline":
            weights = spline(d)
        else:
            tl.device_assert(False, f"Invalid kernel type: {KERNEL}")
    elif NORM == "1":
        if KERNEL == "kaiser_bessel":
            wx = kaiser_bessel(norm1d(dx) / rx, beta)
            wy = kaiser_bessel(norm1d(dy) / ry, beta)
            wz = kaiser_bessel(norm1d(dz) / rz, beta)
            weights = wx[:, None, None] * wy[None, :, None] * wz[None, None, :]
        elif KERNEL == "spline":
            wx = spline(norm1d(dx) / rx)
            wy = spline(norm1d(dy) / ry)
            wz = spline(norm1d(dz) / rz)
            weights = wx[:, None, None] * wy[None, :, None] * wz[None, None, :]
        else:
            tl.device_assert(False, f"Invalid kernel type: {KERNEL}")
    weights_mask = x_mask[:, None, None] & (
        y_mask[None, :, None] & z_mask[None, None, :]
    )
    weights = tl.where(weights_mask, weights, 0.0)
    return weights, x_range, y_range, z_range, x_mask, y_mask, z_mask


@triton.jit
def get_neighborhood(target, kernel_width, base_range):
    lower = target - (kernel_width / 2.0)
    lower = tl.ceil(lower)
    lower = tl.cast(lower, tl.int32)
    return base_range + lower


@triton.jit
def mod_pos(t, n):
    """Modulo but ensures positive return value"""
    return tl.where(t >= 0, t % n, (t % n) + n)



import marimo

__generated_with = "0.12.0"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo
    from jaxtyping import Float
    from torch import Tensor

    from einops import rearrange
    from sigpy import mri
    import torch
    import matplotlib.pyplot as plt
    from torchlinops import Dim
    from torchlinops.utils import cfft2, cifft2
    from torchlinops.linops.nufft import toeplitz_psf, NUFFT
    import numpy as np
    from sigpy.fourier import (
        toeplitz_psf as sp_toeplitz_psf,
        fft as sp_fft,
        ifft as sp_ifft,
    )

    def spiral_2d(
        im_size: tuple,
        n_shots: int = 16,
        alpha: float = 1.5,
        f_sampling: float = 0.4,
        g_max: float = 40.0,
        s_max: float = 100.0,
    ) -> Float[Tensor, "R K D"]:
        """
        Generates an 2-dimensional variable density spiral

        Parameters:
        ----------
        im_size: Tuple
            2D Image resolution tuple
        n_shots : int
            number of phase encodes to cover k-space
        alpha : float
            controls variable density. 1.0 means no variable density, center denisty increases with alpha
        g_max : float
            Maximum gradient amplitude in T/m
        s_max : float
            Maximum gradient slew rate in T/m/s

        Returns:
        ----------
        Float[Tensor, "R K D"]
            k-space trajector with shape (n_shots, n_readout_points,  d), d = len(im_size)
            sigpy scaling - each dim in range [-N//2, N//2]
        """

        # Gen spiral
        trj = mri.spiral(
            fov=1,
            N=max(im_size),
            f_sampling=f_sampling,  # TODO function of self.n_read
            R=1,
            ninterleaves=n_shots,
            alpha=alpha,
            gm=g_max,  # Tesla / m
            sm=s_max,  # Tesla / m / s
        )
        assert trj.shape[0] % n_shots == 0
        trj = rearrange(trj, "(R K) D -> R K D", R=n_shots)
        # Normalize to sigpy scaling
        im_size_2 = np.array(im_size) / 2.0
        trj = trj * im_size_2 / trj.max((0, 1), keepdims=False)
        trj = torch.from_numpy(trj).to(torch.float32)
        return trj

    return (
        Dim,
        Float,
        NUFFT,
        Tensor,
        cfft2,
        cifft2,
        mo,
        mri,
        np,
        plt,
        rearrange,
        sp_fft,
        sp_ifft,
        sp_toeplitz_psf,
        spiral_2d,
        toeplitz_psf,
        torch,
    )


@app.cell
def _(Dim, NUFFT, cifft2, spiral_2d, toeplitz_psf, torch):
    import sigpy as sp

    Nx, Ny = 64, 64
    im_size = (Nx, Ny)

    def make_locs(im_size, mode="cartesian", R=1):
        Nx, Ny = im_size
        if mode == "cartesian":
            kx = torch.linspace(-Nx // 2, Nx // 2 - 1, Nx)
            ky = torch.linspace(-Ny // 2, Ny // 2 - 1, Ny)
            locs = torch.meshgrid(kx, ky, indexing="ij")
            locs = torch.stack(locs, dim=-1)
        elif mode == "random":
            kx = torch.rand(im_size) * (Nx - 1) - Nx // 2
            ky = torch.rand(im_size) * (Ny - 1) - Ny // 2
            locs = torch.stack([kx, ky], dim=-1)
        elif mode == "spiral":
            locs = spiral_2d(im_size)
            locs = locs[::R]
        return locs

    def make_nufft(locs, im_size, oversamp):
        return NUFFT(
            locs, im_size, output_shape=Dim("KxKy"), oversamp=oversamp, width=4.0
        )

    locs = make_locs(im_size, mode="random", R=4)
    # plt.scatter(locs[..., 0], locs[..., 1])
    print(locs)
    nufft = make_nufft(locs.clone(), im_size, oversamp=1.25)
    kern, _, _ = toeplitz_psf(nufft, None)
    otf = kern.weight[0, 0]
    psf = cifft2(otf)
    return (
        Nx,
        Ny,
        im_size,
        kern,
        locs,
        make_locs,
        make_nufft,
        nufft,
        otf,
        psf,
        sp,
    )


@app.cell
def _(otf, plt):
    with plt.rc_context({"figure.dpi": 256}):
        plt.imshow(otf.abs())
    # print(psf.abs()[64:66, 64:66])
    plt.colorbar()
    plt.gcf()
    return


@app.cell
def _(im_size, locs, np, plt, sp_ifft, sp_toeplitz_psf):
    otf_sp = sp_toeplitz_psf(locs.numpy(), im_size)
    psf_sp = sp_ifft(otf_sp)
    with plt.rc_context({"figure.dpi": 256}):
        plt.imshow(np.abs(otf_sp))
        plt.colorbar()
    plt.gcf()
    return otf_sp, psf_sp


@app.cell
def _(im_size, np, psf, psf_sp):
    np.abs(psf_sp[im_size[0] // 2, im_size[1] // 2]) / np.abs(
        psf.numpy()[im_size[0] // 2, im_size[1] // 2]
    )
    return


@app.cell
def _(np, otf_sp):
    np.abs(otf_sp)
    return


@app.cell
def _(NUFFT):
    width = 4
    oversamp = 1.25
    beta = NUFFT.beta(width, oversamp)
    return beta, oversamp, width


@app.cell
def _(NUFFT, beta, oversamp, width):
    img_size = (64, 64)
    oversamp_size = tuple(int(oversamp * s) for s in img_size)
    apod = NUFFT.apodize_weights(img_size, oversamp_size, oversamp, width, beta)
    return apod, img_size, oversamp_size


@app.cell
def _(apod, plt):
    plt.imshow(apod)
    plt.colorbar()
    plt.gcf()
    return


@app.cell
def _(nufft):
    nufft.linops[0].weight[20, 20] * 1000
    return


@app.cell
def _(im_size, nufft, torch):
    delta = torch.zeros(im_size)
    center_idx = tuple(s // 2 for s in im_size)
    delta[center_idx] = 1.0
    (1.0 / nufft(delta)).abs()[center_idx]
    return center_idx, delta


@app.cell
def _(delta, nufft):
    1.0 / nufft(delta)
    return


@app.cell
def _(im_size, np):
    from math import prod

    np.sqrt(prod(im_size))
    return (prod,)


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



from typing import Optional
from jaxtyping import Float
from torch import Tensor

from warnings import warn

import torch.nn as nn

import torchlinops.functional as F
from torchlinops.utils import default_to, default_to_dict

from .namedlinop import NamedLinop
from .nameddim import ELLIPSES, NS, Shape

__all__ = ["Interpolate"]


class Interpolate(NamedLinop):
    def __init__(
        self,
        locs: Float[Tensor, "... D"],
        grid_size: tuple[int, ...],
        batch_shape: Optional[Shape] = None,
        locs_batch_shape: Optional[Shape] = None,
        grid_shape: Optional[Shape] = None,
        # Interp params
        width: float = 4.0,
        kernel: str = "kaiser_bessel",
        norm: str = "1",
        pad_mode: str = "circular",
        kernel_params: Optional[dict] = None,
    ):
        if locs_batch_shape is not None:
            if len(locs_batch_shape) > len(locs.shape) - 1:
                raise ValueError(
                    f"locs_batch_shape has length longer than batch dim of locs. locs_batch_shape: {locs_batch_shape}, locs: {locs.shape}"
                )
        batch_shape = default_to(("...",), batch_shape)
        locs_batch_shape = default_to(("...",), locs_batch_shape)
        grid_shape = default_to(("...",), grid_shape)
        shape = NS(batch_shape) + NS(grid_shape, locs_batch_shape)
        super().__init__(shape)
        self._shape.add("batch_shape", batch_shape)
        self._shape.add("locs_batch_shape", locs_batch_shape)
        self._shape.add("grid_shape", grid_shape)
        self.locs = nn.Parameter(locs, requires_grad=False)
        self.grid_size = grid_size

        # Do this here instead of repeating it in both fn() and adjoint_fn()
        kernel_params = default_to_dict(dict(beta=1.0), kernel_params)
        self._interp_params = {
            "width": width,
            "kernel": kernel,
            "norm": norm,
            "pad_mode": pad_mode,
            "kernel_params": kernel_params,
        }

    def forward(self, x):
        return self.fn(self, x, self.locs)

    @staticmethod
    def fn(linop, x, /, locs):
        return F.interpolate(x, locs, **linop._interp_params)

    @staticmethod
    def adj_fn(linop, x, /, locs):
        return F.interpolate_adjoint(x, locs, linop.grid_size, **linop._interp_params)

    @staticmethod
    def normal_fn(linop, x, /, locs):
        return linop.adj_fn(linop, linop.fn(linop, x, locs), locs)
        # return linop.adj_fn(linop, linop.fn(linop, x, locs), locs)

    def split_forward(self, ibatch, obatch):
        return type(self)(
            self.split_forward_fn(ibatch, obatch, self.locs),
            self.grid_size,
            self._shape.batch_shape,
            self._shape.locs_batch_shape,
            self._shape.grid_shape,
            **self._interp_params,
        )

    def split_forward_fn(self, ibatch, obatch, /, locs):
        """Can only split on locs dimensions"""
        if self._shape.locs_batch_shape == ELLIPSES:
            # warn(f"Attempted to split {self.__class__.__name__} but ")
            return locs

        N = len(self._shape.locs_batch_shape)
        locs_slc = []
        for oslc in obatch[-N:]:
            locs_slc.append(oslc)
        locs_slc.append(slice(None))
        return locs[locs_slc]

    def size(self, dim):
        if dim in self._shape.locs_batch_shape:
            dim_idx = self._shape.locs_batch_shape.index(dim)
            return self.locs.shape[dim_idx]
        elif dim in self._shape.grid_shape:
            dim_idx = self._shape.grid_shape.index(dim)
            return self.grid_size[dim_idx]
        return None



import torch

from torchlinops import Dense, Batch
from torch.profiler import profile, record_function, ProfilerActivity


def main():
    device_idxs = [0, 1]
    devices = [torch.device(f"cuda:{idx}") for idx in device_idxs]
    # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    B, M, N = 6, 100000, 10000
    # Set up
    weight = torch.randn(B, M, N, device=devices[0])
    A = Dense(weight, ("B", "M", "N"), ishape=("N",), oshape=("B", "M"))
    A.to(devices[0])
    A_batch = Batch(A, devices[0], devices[0], torch.float32, torch.float32, B=1)
    # Prepare streams
    # streams = [torch.cuda.Stream(devices[d % len(devices)]) for d in range(B)]
    default_streams = [torch.cuda.Stream(device) for device in devices]
    streams = [default_streams[d % len(devices)] for d in range(B)]
    for linop, stream in zip(A_batch._linops, streams):
        linop.to(stream.device)

    y = torch.empty(B, M, device=devices[0])
    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        on_trace_ready=lambda p: p.export_chrome_trace(f"./trace.json"),
    ) as prof:
        for i in range(10):
            for j, s in enumerate(streams):
                with torch.cuda.stream(s):
                    outputs = A_batch._linops[j](torch.randn(N, device=s.device))
        for j, output in enumerate(outputs):
            y[j] = output.to(y.device)


if __name__ == "__main__":
    main()



from typing import Optional

from copy import deepcopy
import torch.fft as fft

from .namedlinop import NamedLinop
from .nameddim import get_nd_shape, NS, Shape
from .identity import Identity


class FFT(NamedLinop):
    def __init__(
        self,
        ndim: int,
        batch_shape: Optional[Shape] = None,
        grid_shapes: Optional[tuple[Shape, Shape]] = None,
        norm: Optional[str] = "ortho",
        centered: bool = False,
    ):
        """
        Parameters
        ----------
        ndim : int
            Dimension of this fourier transform
        batch_shape: Shape, optional
            Batch dimensions, otherwise defaults to
        grid_shapes: (Shape, Shape), optional
            Pair of shapes corresponding to input (primal domain) and
            output (fourier/dual domain)
            Otherwise, defaults to (Nx[, Ny [, Nz]]) and (Kx[, Ky[, Kz]])
        norm : 'ortho' or None, default 'ortho'
            Normalization to apply to FFT.
            Note: technically only "ortho" results in a "true" forward/adjoint relationship.
        centered : bool, default False
            If True, treats the center of the array (i.e. N // 2) as the origin of k-space and image space
            If False, treats the array index origin (i.e. (0, 0) for 2D) as the origin of k-space and image space
            centered = True mimicks sigpy behavior.




        """
        self.ndim = ndim
        self.dim = tuple(range(-self.ndim, 0))
        self.grid_shapes = grid_shapes
        if grid_shapes is None:
            dim_shape = NS(get_nd_shape(self.dim), get_nd_shape(self.dim, kspace=True))
        else:
            if len(grid_shapes) != 2:
                raise ValueError(
                    f"grid_shapes should consist of two shape tuples but got {grid_shapes}"
                )
            dim_shape = NS(*grid_shapes)
        shape = NS(batch_shape) + dim_shape
        super().__init__(shape)
        self._shape.add("batch_shape", batch_shape)
        self._shape.add("input_grid_shape", grid_shapes[0])
        self._shape.add("output_grid_shape", grid_shapes[1])
        self.norm = norm
        self.centered = centered

    @property
    def batch_shape(self):
        return self._shape.lookup("batch_shape")

    def forward(self, x, /):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x):
        if linop.centered:
            x = fft.ifftshift(x, dim=linop.dim)
        x = fft.fftn(x, dim=linop.dim, norm=linop.norm)
        if linop.centered:
            x = fft.fftshift(x, dim=linop.dim)
        return x

    @staticmethod
    def adj_fn(linop, x):
        if linop.centered:
            x = fft.ifftshift(x, dim=linop.dim)
        x = fft.ifftn(x, dim=linop.dim, norm=linop.norm)
        if linop.centered:
            x = fft.fftshift(x, dim=linop.dim)
        return x

    @staticmethod
    def normal_fn(linop, x):
        return x

    def split_forward(self, ibatch, obatch):
        new = type(self)(
            self.ndim,
            self.batch_shape,
            self.grid_shapes,
            self.norm,
            self.centered,
        )
        new._shape = deepcopy(self._shape)
        return new

    def normal(self, inner=None):
        if inner is None:
            return Identity(self.ishape)
        return super().normal(inner)



import pytest

from math import prod, ceil, floor

import torch
import sigpy as sp

from torchlinops.functional import ungrid
from torchlinops.utils import to_pytorch, from_pytorch
from torchlinops.functional._interp.tests._valid_pts import get_valid_locs

PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


# TODO add kernel type
@pytest.mark.parametrize("kernel_type", ["kaiser_bessel", "spline"])
@pytest.mark.parametrize("padding_mode", ["zero", "circular"])
@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["real", "complex"])
@pytest.mark.parametrize(
    "spec",
    [
        "small1d",
        "medium1d",
        pytest.param("large1d", marks=PYTEST_GPU_MARKS),
        "tiny2d",
        "small2d",
        pytest.param("medium2d", marks=PYTEST_GPU_MARKS),
        pytest.param("large2d", marks=PYTEST_GPU_MARKS),
        "small3d",
        pytest.param("medium3d", marks=PYTEST_GPU_MARKS),
        pytest.param("large3d", marks=PYTEST_GPU_MARKS + [pytest.mark.slow]),
    ],
)
def test_ungrid(kernel_type, padding_mode, dev, dtype, spec, request):
    device = torch.device(dev)
    dtype = torch.complex64 if dtype == "complex" else torch.float32
    spec_name = spec
    spec = request.getfixturevalue(spec)

    width = spec["width"]
    grid_size = spec["grid_size"]
    locs_batch_size = spec["locs_batch_size"]
    ndim = len(grid_size)
    npts = prod(locs_batch_size)
    ishape = (*spec["N"], *grid_size)
    if spec_name == "large3d":
        vals = torch.randn(ishape, dtype=dtype, device=device) + 4.0
    else:
        vals = torch.arange(prod(ishape)).reshape(ishape).to(dtype).to(device)
    # locs = tuple(
    #     torch.rand(spec["npts"], device=device) + (w / 2 - 1) for d in range(ndim)
    # )
    # locs = torch.stack(locs, dim=-1).contiguous()
    if padding_mode == "zero":
        locs = get_valid_locs(locs_batch_size, grid_size, ndim, width, device)
    elif padding_mode == "circular":
        locs = get_valid_locs(
            locs_batch_size, grid_size, ndim, width, device, valid=False
        )

    interp = ungrid(
        vals,
        locs,
        width=width,
        kernel=kernel_type,
        pad_mode=padding_mode,
    )

    # Test against sigpy
    vals = from_pytorch(vals)
    # locs = from_pytorch(locs.flip(dims=(-1,)).contiguous())
    locs = from_pytorch(locs)

    # Sigpy quirks:
    # - Circular padding
    # - weight is product of 1D kernel evals along each coordinate axis
    interp_sp = sp.interpolate(vals, locs, kernel=kernel_type, width=width)
    interp_sp = to_pytorch(interp_sp)

    assert torch.allclose(interp, interp_sp)



from typing import Optional
from copy import copy, deepcopy

import torch.nn.functional as F

from .namedlinop import NamedLinop
from .identity import Identity
from .nameddim import ND, get_nd_shape, NS, NDorStr, Shape
from torchlinops.utils import default_to


__all__ = ["PadLast"]


class PadLast(NamedLinop):
    """Zero Pad the last dimensions of the input volume
    Padding is centered:
    - TODO: support non-centered padding?
    ishape: [B... Nx Ny [Nz]]
    oshape: [B... Nx1 Ny1 [Nz1]]

    """

    def __init__(
        self,
        pad_im_size: tuple[int, ...],
        im_size: tuple[int, ...],
        in_shape: Optional[Shape] = None,
        out_shape: Optional[Shape] = None,
        batch_shape: Optional[Shape] = None,
    ):
        if len(pad_im_size) != len(im_size):
            raise ValueError(
                f"Padded and unpadded dims should be the same length. padded: {pad_im_size} unpadded: {im_size}"
            )

        if in_shape is None:
            self.in_im_shape = ND.infer(get_nd_shape(im_size))
        else:
            self.in_im_shape = ND.infer(in_shape)
        if out_shape is None:
            self.out_im_shape = tuple(
                d.next_unused(self.in_im_shape) for d in self.in_im_shape
            )
        else:
            self.out_im_shape = out_shape
        batch_shape = default_to(("...",), batch_shape)

        shape = NS(batch_shape) + NS(self.in_im_shape, self.out_im_shape)
        super().__init__(shape)
        self.D = len(im_size)
        self.im_size = tuple(im_size)
        self.pad_im_size = tuple(pad_im_size)
        self.in_im_size = tuple(im_size)
        self.out_im_size = tuple(pad_im_size)
        # for psz in pad_im_size:
        #     assert not (psz % 2), "Pad sizes must be even"

        # sizes = [
        #     [(psz - isz) // 2] * 2
        #     for psz, isz in zip(self.out_im_size, self.in_im_size)
        # ]
        # self.pad = sum(sizes, start=[])
        # self.pad.reverse()

        self.pad = pad_to_size(self.im_size, self.pad_im_size)

        # Make crop slice that undoes padding
        # Need to reverse crop_slice because padding is reversed
        self.crop_slice = crop_slice_from_pad(self.pad)

    def forward(self, x):
        """Pad the last n dimensions of x"""
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x, /):
        if tuple(x.shape[-linop.D :]) != linop.im_size:
            raise ValueError(
                f"Mismatched shapes: expected {linop.im_size} but got {x.shape[-linop.D :]}"
            )
        pad = linop.pad + [0, 0] * (x.ndim - linop.D)
        return F.pad(x, pad)

    @staticmethod
    def adj_fn(linop, y, /):
        """Crop the last n dimensions of y"""
        if tuple(y.shape[-linop.D :]) != linop.pad_im_size:
            raise ValueError(
                f"Mismatched shapes: expected {linop.pad_im_size} but got {y.shape[-linop.D :]}"
            )
        slc = [slice(None)] * (y.ndim - linop.D) + linop.crop_slice
        return y[slc]

    def adjoint(self):
        adj = super().adjoint()
        adj.in_im_shape, adj.out_im_shape = self.out_im_shape, self.in_im_shape
        adj.in_im_size, adj.out_im_size = self.out_im_size, self.in_im_size
        return adj

    def split_forward(self, ibatch, obatch):
        self.split_forward_fn(ibatch, obatch)
        return self

    def split_forward_fn(self, ibatch, obatch, /):
        for islc, oslc in zip(ibatch[-self.D :], obatch[-self.D :]):
            if islc != slice(None) or oslc != slice(None):
                raise ValueError(
                    f"{type(self).__name__} cannot be split along image dim"
                )
        return None

    def size(self, dim: str):
        return self.size_fn(dim)

    def size_fn(self, dim: str, /):
        if dim in self.ishape[-self.D :]:
            return self.in_im_size[self.in_im_shape.index(dim)]
        elif dim in self.oshape[-self.D :]:
            return self.out_im_size[self.out_im_shape.index(dim)]
        return None


def pad_to_scale(grid_size, scale_factor):
    """Convenience wrapper for pad_to_size but with a scale factor instead"""
    padded_size = [int(i * scale_factor) for i in grid_size]
    return pad_to_size(grid_size, padded_size)


def pad_to_size(grid_size, padded_size):
    """Construct a padding list suitable for torch.nn.functional.pad

    Pad will take an input with size `grid_size` and return an output with size `factor * grid_size`

    Padding rules:
    odd/even image + even pad -> [pad // 2, pad // 2]
    even image + odd pad -> [pad // 2, pad // 2 + 1]
    odd image + odd pad -> [pad // 2 + 1, pad // 2]

    Preserves the "center" of the image or kspace if it has been fftshifted

    Useful for e.g. padding an image to increase resolution in fourier domain
    """
    if len(grid_size) != len(padded_size):
        raise ValueError(
            f"Dimension mismatch: cannot pad from size {grid_size} to size {padded_size}."
        )
    total_padding = [p - i for p, i in zip(padded_size, grid_size)]
    pad = []
    for i, tp in zip(grid_size, total_padding):
        if tp % 2:  # pad is odd
            if i % 2:  # im is odd
                pad_left = tp // 2 + 1
                pad_right = tp // 2
            else:  # im is even
                pad_left = tp // 2
                pad_right = tp // 2 + 1
        else:
            pad_left = tp // 2
            pad_right = tp // 2
        pad.append([pad_left, pad_right])
    pad.reverse()
    return sum(pad, start=[])


def crop_slice_from_pad(pad):
    """From a padding list, get the corresponding slicing that undoes it."""
    crop_slice = []
    for i in range(len(pad) // 2):
        start = pad[2 * i]
        stop = -pad[2 * i + 1]
        if stop == 0:
            # Don't set stop to 0, set it to the end
            stop = None
        crop_slice.append(slice(start, stop))
    crop_slice.reverse()
    return crop_slice



from torch import Tensor

from itertools import cycle

import torch
from tqdm import tqdm

from .batch import Batch

__all__ = ["MPBatch"]


class MPBatch(Batch):
    def __init__(self, *args, devices: list[torch.device], **kwargs):
        """
        Parameters
        ----------
        Same as Batch.

        Additional parameters:
        devices : list[torch.device]
            A list of devices to parallelize the linop over
            MPBatch will attempt to parallelize the work evenly over
            all the devices specified.


        """
        self.devices = devices
        super().__init__(*args, **kwargs)

    def forward(self, x: Tensor):
        """
        Computes on stated devices but ultimately returns on the output device
        """
        # Complete the size specifications
        for dim, total in zip(self.ishape, x.shape):
            self.sizes[dim] = total

        # For holding input and output batches
        xs = []
        ys = []

        # Distribute linops and inputs
        for linop, in_batch, device in zip(
            self._linops, self._input_batches, cycle(self.devices)
        ):
            linop.to(device)
            xs.append(x[in_batch].to(device))

        # Run linops on inputs
        for linop, xbatch in tqdm(
            zip(self._linops, xs),
            total=len(self._linops),
            desc=f"MPBatch({self.repr_name}: {self.batch_sizes})",
            disable=(not self.pbar),
        ):
            ybatch = linop(xbatch)
            ys.append(ybatch)

        # Gather outputs
        y = torch.zeros(
            tuple(self.sizes[dim] for dim in self.oshape),
            dtype=self.output_dtype,
            device=self.output_device,
        )
        for ybatch, out_batch in zip(ys, self._output_batches):
            y[out_batch] += ybatch.to(self.output_device)
        return y

    def adjoint(self):
        batch_sizes = {str(k): v for k, v in self.batch_sizes.items()}
        adj = type(self)(
            linop=self.linop.H,
            input_device=self.output_device,
            output_device=self.input_device,
            input_dtype=self.output_dtype,
            output_dtype=self.input_dtype,
            name=self.name + ".H",
            pbar=self.pbar,
            devices=self.devices,
            post_batch_hook=self.post_batch_hook,
            **batch_sizes,
        )
        return adj

    def normal(self, inner=None):
        normal_linop = self.linop.N
        # Collect shape updates from computing the normal
        shape_updates = getattr(normal_linop, "_shape_updates", {})
        for d, nd in shape_updates.items():
            if d in self.batch_sizes:
                self.batch_sizes[shape_updates[d]] = self.batch_sizes[d]
        batch_size_kwargs = {str(k): v for k, v in self.batch_sizes.items()}
        normal = type(self)(
            linop=self.linop.N,
            input_device=self.input_device,
            output_device=self.input_device,
            input_dtype=self.input_dtype,
            output_dtype=self.input_dtype,
            name=self.name + ".N",
            pbar=self.pbar,
            devices=self.devices,
            post_batch_hook=self.post_batch_hook,
            **batch_size_kwargs,
        )
        return normal



from typing import Callable, Tuple, Optional
from torch import Tensor

import torch
from tqdm import tqdm

from torchlinops.utils import default_to_dict


__all__ = ["power_method"]


def power_method(
    A: Callable[[Tensor], Tensor],
    ishape: Tuple,
    v_init: Optional[Tensor] = None,
    max_iters: int = 50,
    device: torch.device = "cpu",
    eps: float = 0.0,
    tol: float = 1e-5,
    dim: Optional[int | Tuple] = None,
    tqdm_kwargs: Optional[dict] = None,
) -> tuple[Tensor, Tensor]:
    """Finds the maximum eigenvalue (in absolute value) of square matrix A

    Parameters
    ----------
    dim : Optional[int | Tuple]
        If not None, compute eigenvalues along only that dimension
        Enables batched power method over several stacked matrices

    Returns
    -------
    Tensor : The eigenvector
    Tensor : its associated eigenvalue

    """
    # Default values
    tqdm_kwargs = default_to_dict(dict(desc="Power Method"), tqdm_kwargs)
    if v_init is None:
        v = torch.randn(ishape, dtype=torch.complex64, device=device)
    else:
        v = v_init.clone()

    # Initialize
    A.to(device)
    vnorm = torch.linalg.vector_norm(v, dim=dim, keepdim=True)
    v = v / (vnorm + eps)
    pbar = tqdm(range(max_iters), total=max_iters, **tqdm_kwargs)
    for _ in pbar:
        vnorm_old = vnorm.clone()
        v = A(v)
        vnorm = torch.linalg.vector_norm(v, dim=dim, keepdim=True)
        v = v / (vnorm + eps)
        rdiff = (torch.abs(vnorm_old - vnorm) / torch.abs(vnorm_old)).max()

        # Display progress
        postfix = {"rdiff": rdiff.item()}
        if dim is None:
            postfix["e_val"] = vnorm.item()
        pbar.set_postfix(postfix)
        if rdiff < tol:
            break
    return v, vnorm.squeeze()



import pytest

import torch
import numpy as np

from torchlinops.functional._interp.tests._valid_pts import get_valid_locs
from torchlinops import NUFFT, Diagonal, Dense, Dim
from torchlinops.utils import from_pytorch

from torchlinops.linops.nufft import toeplitz_psf

from sigpy.fourier import toeplitz_psf as sp_toeplitz_psf


@pytest.fixture
def nufft_params():
    width = 4.0
    oversamp = 1.25
    # grid_size = (120, 119, 146)
    grid_size = (64, 64, 64)
    padded_size = [int(i * oversamp) for i in grid_size]
    locs = get_valid_locs(
        (20, 500),
        grid_size,
        len(grid_size),
        width,
        "cpu",
        centered=True,
    )
    return {
        "width": width,
        "oversamp": oversamp,
        "grid_size": grid_size,
        "padded_size": padded_size,
        "locs": locs,
    }


@pytest.fixture
def nufft_linop(nufft_params):
    locs = nufft_params["locs"]
    grid_size = nufft_params["grid_size"]
    width = nufft_params["width"]
    oversamp = nufft_params["oversamp"]
    linop = NUFFT(
        locs.clone(),
        grid_size,
        output_shape=Dim("RK"),
        batch_shape=Dim("A"),
        width=width,
        oversamp=oversamp,
    )
    return linop


@pytest.fixture
def simple_inner(nufft_params):
    locs = nufft_params["locs"]
    weight = torch.randn(locs.shape[:-1])
    linop = Diagonal(weight, ioshape=Dim("ARK"), broadcast_dims=Dim("A"))
    return linop


@pytest.fixture
def dense_inner(nufft_params):
    A = 2
    weight = torch.randn(A, A, dtype=torch.complex64)
    linop = Dense(
        weight,
        weightshape=Dim("AA1"),
        ishape=Dim("ARK"),
        oshape=Dim("A1RK"),
    )
    return linop


@pytest.mark.parametrize("inner_type", ["simple_inner", "dense_inner", None])
def test_toeplitz_full(inner_type, nufft_linop, nufft_params, request):
    if inner_type is not None:
        inner = request.getfixturevalue(inner_type)
    else:
        inner = None
    kernel = toeplitz_psf(nufft_linop, inner)

    # Test against sigpy for no inner only
    if inner_type is None:
        psf = kernel.weight
        coord = from_pytorch(nufft_params["locs"].clone())
        psf_sp = sp_toeplitz_psf(
            coord,
            nufft_linop.grid_size,
            oversamp=nufft_linop.oversamp,
            width=nufft_linop.width,
        )
        assert np.isclose(psf_sp, psf.numpy(), rtol=1e-1).sum() / psf_sp.size > 0.99



import pytest
from math import prod
import torch

from torchlinops.functional import unfold

PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["real", "complex"])
@pytest.mark.parametrize(
    "spec",
    [
        "small1d",
        # "medium1d",
        # pytest.param("large1d", marks=PYTEST_GPU_MARKS),
        "small2d",
        # pytest.param("medium2d", marks=PYTEST_GPU_MARKS),
        # pytest.param("large2d", marks=PYTEST_GPU_MARKS),
    ],
)
def test_unfold_mask(dev, dtype, spec, request):
    spec = request.getfixturevalue(spec)
    device = torch.device(dev)
    dtype = torch.complex64 if dtype == "complex" else torch.float32

    ishape = (*spec["N"], *spec["shape"])
    x = torch.arange(prod(ishape)).reshape(ishape)
    # x = torch.ones(prod(ishape)).reshape(ishape)
    x = x.to(device).to(dtype)

    mask = spec["mask"]
    y_th = unfold(x, spec["block_size"], spec["stride"])
    y_th_masked = unfold(x, spec["block_size"], spec["stride"], mask)

    assert torch.allclose(y_th[..., mask.to(device)], y_th_masked)



import pytest

import torch
from torchlinops import NamedLinop, NS


def test_namedlinop_adjoint():
    A = NamedLinop(NS(("A", "B"), ("C",)))
    AH = A.H


def test_namedlinop_split():
    A = NamedLinop(NS(("A", "B"), ("C",)))
    A1 = A.split(A, {})


def test_namedlinop_normal():
    A = NamedLinop(NS(("A", "B"), ("C",)))
    AN = A.N


def test_namedlinop_chain_normal_split():
    A = NamedLinop(NS(("A", "B"), ("C",)))
    B = NamedLinop(NS(("C",), ("C1",)))
    AB = B @ A
    ABN = AB.N


def test_namedlinop_custom_name():
    A = NamedLinop(NS(("A", "B"), ("C",)))
    assert str(A).startswith("NamedLinop")

    custom_name = "Test"
    A = NamedLinop(NS(("A", "B"), ("C",)), name=custom_name)
    assert str(A).startswith(custom_name)


def test_namedlinop_matrix_vector_mul():
    A = NamedLinop(NS(("A", "B"), ("C",)))
    x = torch.tensor(3)
    assert A(x).allclose(A @ x)



import torch

from torchlinops import NS
from torchlinops import (
    NamedLinop,
    Chain,
    Dense,
    Diagonal,
    FFT,
    Scalar,
    Identity,
    Add,
)


def test_dense():
    M, N = 9, 3
    weight = torch.randn(M, N, dtype=torch.complex64)
    weightshape = ("M", "N")
    x = torch.randn(N, dtype=torch.complex64)
    ishape = ("N",)
    # y = torch.randn(M)
    oshape = ("M",)
    A = Dense(weight, weightshape, ishape, oshape)
    assert torch.isclose(A.N(x), A.H(A(x))).all()
    # Make sure dense's normal doesn't create a chain (unnecessary)
    # If desired, just make the linop explicitly
    assert not isinstance(A.N, Chain)
    assert A.N.ishape == ("N",)
    assert A.N.oshape == ("N1",)


def test_diagonal_normal():
    M = 10
    N, P = 5, 7
    weight = torch.randn(M, 1, 1, dtype=torch.complex64)
    # weightshape = ("M",)
    x = torch.randn(M, N, P, dtype=torch.complex64)
    ioshape = ("M", "N", "P")
    A = Diagonal(weight, ioshape)
    assert torch.isclose(A.N(x), A.H(A(x))).all()
    assert not isinstance(A.N, Chain)



import pytest

from torchlinops import NamedDimCollection, ANY


def test_overwrite_any():
    old_shape = ("A", ANY, "C", ANY)
    new_shape = ("A", "B", "C", "D")
    ndc = NamedDimCollection(shape=old_shape)
    ndc.update("shape", new_shape)
    assert ndc.shape == new_shape



import torch

from torchlinops import Dense, Batch
from torch.profiler import profile, record_function, ProfilerActivity

from _parallel import parallel_call


def main():
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    B, M, N = 3, 100000, 1000
    # Set up
    weight = torch.randn(B, M, N, device=device)
    A = Dense(weight, ("B", "M", "N"), ishape=("N",), oshape=("B", "M"))
    A.to(device)
    A_batch = Batch(A, device, device, torch.float32, torch.float32, B=1)
    A_batch = torch.compile(A_batch)

    x = torch.randn(N, device=device)
    # y = torch.empty(B, M, device=device)
    # with profile(
    #     activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
    #     on_trace_ready=lambda p: p.export_chrome_trace(f"./trace.json"),
    # ) as prof:
    y = parallel_call(A_batch._linops, [(x,) for _ in range(B)], do_profile=True)


if __name__ == "__main__":
    main()



import torch
from math import prod

from fold import fold
import sigpy as sp
import cupy as cp
import numpy as np

from utils.benchmark import benchmark
from utils import Indenter, device_ordinal
from utils import from_pytorch, to_pytorch

from nblocks import get_nblocks


def main():
    device = torch.device("cuda:0")
    N = 2
    Nx = 100

    im_size = (Nx, Nx)

    block_dim = (4, 4)
    stride = (2, 2)

    nblocks = get_nblocks(im_size, block_dim, stride)
    # mask = torch.zeros(*block_dim, dtype=bool)
    # mask[::2] = 1
    mask = None

    dtype = torch.complex64
    cp_dtype = np.float32

    ### Triton Version ###
    def random_fold_triton():
        x = torch.arange(
            N * prod(nblocks) * prod(block_dim), dtype=torch.float32, device=device
        ).reshape(N, *nblocks, *block_dim)
        # x = torch.randn(N * Nx, dtype=torch.complex64, device=device).reshape(N, Nx)
        # x = torch.randn((N, Nx), device=device)
        return fold(x, im_size, block_dim, stride, mask)

    triton_res, _ = benchmark(random_fold_triton, num_iters=100)
    summarize(triton_res, "triton")

    ### torch.compile Version ###
    # Mildly slower... also takes forever to compile
    # def random_unfold_torch():
    #     x = torch.randn((N, Nx), device=device)
    #     return unfold_torch(x, block_dim, stride)

    # torch_res, _ = benchmark(random_unfold_torch)
    # summarize(torch_res, "torch.compile")

    ### Cupy Version ###
    dev = sp.Device(device_ordinal(device))

    def random_fold_sp():
        # x = torch.randn((N, Nx), device=device)
        # x = from_pytorch(x)
        xp = dev.xp
        with dev:
            x = xp.random.randn(
                N, *nblocks, *block_dim, dtype=cp_dtype
            ) + 1j * xp.random.randn(N, *nblocks, *block_dim, dtype=cp_dtype)
            return sp.blocks_to_array(x, (N, *im_size), block_dim, stride)

    sp_res, _ = benchmark(random_fold_sp, num_iters=100)
    summarize(sp_res, "sp")

    # Test correctness
    x = torch.randn((N, *nblocks, *block_dim), device=device)
    Bx_triton = fold(x, im_size, block_dim, stride, mask)
    Bx_sp = sp.blocks_to_array(from_pytorch(x), (N, *im_size), block_dim, stride)
    assert torch.allclose(Bx_triton, to_pytorch(Bx_sp), rtol=1e-3)


def summarize(benchmark_result, name: str):
    with Indenter() as indent:
        print(name)
        with indent:
            indent.print(
                f"Mean Time: {np.mean(benchmark_result['timings_ms']):0.3f} ms"
            )
            indent.print(f"Min Time: {np.min(benchmark_result['timings_ms']):0.3f} ms")
            indent.print(f"Max Time: {np.max(benchmark_result['timings_ms']):0.3f} ms")
            indent.print(f"Memory: {benchmark_result['max_mem_bytes']} bytes")


if __name__ == "__main__":
    main()



from math import prod

import pytest
import torch
from torchlinops.functional import fold, unfold
from torchlinops.functional._unfold.nblocks import get_nblocks
from torchlinops.utils import cdata


PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


@pytest.fixture
def medium2d():
    N = (2, 2)
    shape = (33, 43)
    block_size = (7, 7)
    stride = (2, 2)
    return N, shape, block_size, stride


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
def test_fold_allocated_buffer(dev, medium2d):
    device = torch.device(dev)
    dtype = torch.float32
    N, shape, block_size, stride = medium2d

    nblocks = get_nblocks(shape, block_size, stride)
    ishape = (*N, *nblocks, *block_size)
    oshape = (*N, *shape)

    x = torch.arange(prod(ishape)).reshape(ishape)
    x = x.to(device).to(dtype)
    # TODO: why does this not fail??
    y = fold(x, shape, block_size, stride)

    # Do it again, but with output allocated
    x2 = torch.randn_like(x)
    y2 = fold(x2, shape, block_size, stride, output=y)
    y3 = fold(x2, shape, block_size, stride)  # buffer not reused

    assert cdata(y) == cdata(y2)

    assert cdata(y) == cdata(y2)
    assert cdata(y3) != cdata(y2)  # Buffer should not be shared
    assert (y3 == y2).all()  # However, results should match


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
def test_unfold_allocated_buffer(dev, medium2d):
    device = torch.device(dev)
    dtype = torch.float32
    N, shape, block_size, stride = medium2d

    nblocks = get_nblocks(shape, block_size, stride)
    ishape = (*N, *shape)

    x = torch.arange(prod(ishape)).reshape(ishape)
    x = x.to(device).to(dtype)
    y = unfold(x, block_size, stride)

    # Do it again, but with output allocated
    x2 = torch.randn_like(x)
    y2 = unfold(x2, block_size, stride, output=y)
    y3 = unfold(x2, block_size, stride)  # Buffer not reused

    assert cdata(y) == cdata(y2)
    assert cdata(y3) != cdata(y2)  # Buffer should not be shared
    assert (y3 == y2).all()  # But results should match



import json

import torch
import torch.nn as nn
import torch.multiprocessing as mp
from torch.profiler import profile, record_function, ProfilerActivity

try:
    mp.set_start_method("forkserver")
except RuntimeError:
    pass

__all__ = ["parallel_call", "merge_traces"]


def parallel_call(
    modules: list[nn.Module],
    args: list[tuple],
    do_profile: bool = False,
):
    """Run a list of modules on a list of arguments in parallel

    Note: if the modules have any parameters that require gradient, this might not work.

    """
    if len(modules) != len(args):
        raise ValueError(
            f"modules and args must be same length but got {len(modules) } != {len(args)}"
        )
    # Initialize input queues, output_queues, and received queues
    n_procs = len(modules)
    qins = [mp.Queue() for _ in range(n_procs)]
    qouts = [mp.Queue() for _ in range(n_procs)]
    recvs = [mp.Event() for _ in range(n_procs)]

    # Start processes and send data
    procs = []
    for rank, mod, args, qin, qout, recv in zip(
        range(n_procs), modules, args, qins, qouts, recvs
    ):
        proc = mp.Process(target=call_module, args=(rank, qin, qout, recv, do_profile))
        proc.start()
        qin.put((mod, args))
        procs.append(proc)

    # Receive data, signal child processes to exit,and join processes
    outputs = []
    for proc, qout, recv in zip(procs, qouts, recvs):
        outputs.append(qout.get())
        recv.set()
        proc.join()

    return outputs


def call_module(
    rank: int,
    qin: mp.Queue,
    qout: mp.Queue,
    recv: mp.Event,
    do_profile: bool = False,
):
    """Helper function for parallel call

    rank : int
        The id of this process

    """
    # Get module and args from input queue
    mod, args = qin.get()

    # Run module on args
    if do_profile:
        with profile(
            activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
            on_trace_ready=lambda p: p.export_chrome_trace(
                f"./worker_{rank}_trace.json"
            ),
        ) as prof:
            out = mod(*args)
    else:
        out = mod(*args)

    # Put output in the output queue
    qout.put(out)

    # Wait for parent process to signal received
    recv.wait()


def merge_traces(input_files, output_file):
    """

    Example usage
    trace_files = glob.glob("./worker_*_trace.json")
    merge_traces(trace_files, "merged_trace.json")
    """
    all_traces = {}
    all_traces["traceEvents"] = []

    for i, trace_file in enumerate(input_files):
        with open(trace_file, "r") as f:
            trace_data = json.load(f)
            events = trace_data["traceEvents"]
            for event in events:
                event["pid"] = str(event["pid"]) + f"_{i}"

            # Important step
            all_traces["traceEvents"].extend(events)
            # Less important steps
            all_traces["schemaVersion"] = trace_data["schemaVersion"]
            all_traces["deviceProperties"] = trace_data["deviceProperties"]
    all_traces["traceName"] = str(output_file)

    # Write combined traces to a new file
    with open(output_file, "w") as f:
        json.dump(all_traces, f)


### Testing ###
class TestAdd(nn.Module):
    """Needs to be defined at top level of a module"""

    def __init__(self, bias):
        super().__init__()
        self.bias = nn.Parameter(bias, requires_grad=False)

    def forward(self, a):
        def flipsign(x):
            """Test functions defined within methods"""
            return -x

        for i in range(1000):
            flipsign(a + self.bias)
        return flipsign(a + self.bias)


if __name__ == "__main__":
    n_parallel = 3
    devs = [torch.device(f"cuda:{i}") for i in range(n_parallel)]
    modules = [
        TestAdd(torch.randn(i + 1, 1, device=devs[i])) for i in range(n_parallel)
    ]
    args = [(torch.randn(i + 1, device=devs[i]),) for i in range(n_parallel)]
    outputs = parallel_call(modules, args, do_profile=True)
    print([f"{t.shape} : {t.device}" for t in outputs])
    import glob, json

    trace_files = glob.glob("./worker_*_trace.json")
    print(trace_files)
    merge_traces(trace_files, "merged_trace.json")

    breakpoint()



import pytest

import torch

from torchlinops import Concat, Dense, Diagonal, ND
from einops import einsum

from torchlinops.tests.test_base import BaseNamedLinopTests


def test_concat_init():
    N = 1
    P, Q = 3, 4
    ishape = ("N", "Q")
    oshape = ("N", "P")

    # Diagonal stacking
    A = Dense(torch.randn(N, P, Q), ("N", "P", "Q"), ishape, oshape)
    B = Dense(torch.randn(N, P, Q), ("N", "P", "Q"), ishape, oshape)
    C = Concat(A, B, idim="N")
    C2 = Concat(A, B, idim="Q")
    C3 = Concat(A, B, idim="Q", odim="P")


def test_concat_shape_inference():
    shape = (ND("A"), ND("B"), ND("..."), ND("C"), ND("D"), ND("E"))

    d1 = Concat._infer_dim_idx("B", shape)
    assert d1 == 1
    d2 = Concat._infer_dim_idx("C", shape)
    assert d2 == -3


@pytest.fixture
def denselinops():
    N = 1
    P, Q = 3, 4
    ishape = ("N", "Q")
    oshape = ("N", "P")
    wA = torch.randn(N, P, Q)
    wB = torch.randn(N, P, Q)
    A = Dense(wA, ("N", "P", "Q"), ishape, oshape)
    B = Dense(wB, ("N", "P", "Q"), ishape, oshape)
    return A, B


def test_concat_horizontal(denselinops):
    A, B = denselinops
    # Horizontal stack
    C = Concat(A, B, idim="Q")
    assert C.size("Q") == A.size("Q") + B.size("Q")
    x = torch.randn(C.size("N"), C.size("Q"))
    Cx = C(x)
    xs = x.tensor_split(C.islices, dim=C.idim_idx)[:-1]
    Cx_ref = A(xs[0]) + B(xs[1])
    assert Cx.allclose(Cx_ref)


def test_concat_vertical(denselinops):
    A, B = denselinops
    # Vertical stack
    C = Concat(A, B, odim="P")
    assert C.size("P") == A.size("P") + B.size("P")
    x = torch.randn(C.size("N"), C.size("Q"))
    Cx = C(x)
    Cx_ref = torch.concatenate((A(x), B(x)), dim=-1)
    assert Cx.allclose(Cx_ref)


def test_concat_diagonal(denselinops):
    A, B = denselinops
    # Diagonal stack
    C = Concat(A, B, idim="Q", odim="P")
    assert C.size("Q") == A.size("Q") + B.size("Q")
    assert C.size("P") == A.size("P") + B.size("P")
    x = torch.randn(C.size("N"), C.size("Q"))
    Cx = C(x)
    xs = x.tensor_split(C.islices, dim=C.idim_idx)[:-1]
    Cx_ref = torch.concatenate((A(xs[0]), B(xs[1])), dim=-1)
    assert Cx.allclose(Cx_ref)


def test_concat_split(denselinops):
    A, B = denselinops
    # Vertical stack
    C = Concat(A, B, odim="P")

    # Single Linop
    C1 = C.split(C, {"P": slice(0, 3)})
    assert isinstance(C1, Dense)
    assert (C1.weight == A.weight).all()

    # Multiple linops (overlapping)
    C2 = C.split(C, {"P": slice(2, 6)})
    assert isinstance(C2, Concat)
    assert (C2[0].weight == A.weight[:, 2:, :]).all()
    assert (C2[1].weight == B.weight[:, :, :]).all()

    # Horizontal stack
    C = Concat(A, B, idim="Q")
    # Single Linop
    C3 = C.split(C, {"Q": slice(0, 4)})
    assert isinstance(C3, Dense)
    assert (C3.weight == A.weight).all()

    # Multiple linops (overlapping)
    C4 = C.split(C, {"Q": slice(2, 6)})
    assert isinstance(C4, Concat)
    assert (C4[0].weight == A.weight[:, :, 2:]).all()
    assert (C4[1].weight == B.weight[:, :, :2]).all()



from math import prod, ceil

import torch
import sigpy as sp

from torchlinops.utils.benchmark import benchmark_and_summarize
from torchlinops.utils._device import device_ordinal
from torchlinops.utils import to_pytorch, from_pytorch
from torchlinops.functional import grid

from torchlinops.functional._interp.tests._valid_pts import get_valid_locs


def main():
    device = torch.device("cuda")
    dtype = torch.complex64
    N = (1,)
    grid_size = (120,)
    npts = (75,)
    width = 4.0
    kernel = "kaiser_bessel"

    ishape = (*N, *npts)

    ndim = len(grid_size)
    # grid_size = torch.tensor(grid_size, device=device)

    def gen_random_triton():
        # vals = torch.arange(prod(ishape)).reshape(ishape).to(dtype).to(device)
        vals = torch.randn(ishape, dtype=dtype, device=device)
        locs = get_valid_locs(npts, grid_size, ndim, width, device)
        return

    def grid_random_triton():
        # vals = torch.arange(prod(ishape)).reshape(ishape).to(dtype).to(device)
        vals = torch.randn(ishape, dtype=dtype, device=device)
        locs = get_valid_locs(npts, grid_size, ndim, width, device)
        out = grid(vals, locs, grid_size, width, kernel)
        return out

    benchmark_and_summarize(
        gen_random_triton,
        num_iters=100,
        name="torch gen random",
        ignore_first=10,
    )

    res_triton, out_triton = benchmark_and_summarize(
        grid_random_triton,
        num_iters=100,
        name="triton",
        ignore_first=10,
    )

    dev = sp.Device(device_ordinal(device))

    def gen_random_sigpy():
        xp = dev.xp
        with dev:
            if dtype == torch.complex64:
                vals = xp.random.randn(*ishape, dtype=float) + 1j * xp.random.randn(
                    *ishape, dtype=float
                )
            else:
                vals = xp.random.randn(N, Nx, dtype=cp_dtype)
            locs = from_pytorch(get_valid_locs(npts, grid_size, ndim, width, device))

        return

    def grid_random_sigpy():
        xp = dev.xp
        with dev:
            if dtype == torch.complex64:
                vals = xp.random.randn(*ishape, dtype=float) + 1j * xp.random.randn(
                    *ishape, dtype=float
                )
            else:
                vals = xp.random.randn(N, Nx, dtype=cp_dtype)
            locs = from_pytorch(get_valid_locs(npts, grid_size, ndim, width, device))
            out = sp.gridding(vals, locs, grid_size, kernel=kernel, width=width)
        return out

    benchmark_and_summarize(
        gen_random_sigpy,
        num_iters=100,
        name="sigpy gen random",
        backend="cupy",
        ignore_first=10,
    )

    res_sigpy, out_sigpy = benchmark_and_summarize(
        grid_random_sigpy,
        num_iters=100,
        name="sigpy",
        backend="cupy",
        ignore_first=10,
    )


if __name__ == "__main__":
    main()



import inspect

__all__ = ["check_signature"]


def check_signature(signature_spec: list, *args, **kwargs):
    """Check if the provided arguments match the specified signature.

    Parameters
    ----------
    signature_spec : list
        A list that defines the function signature.
    *args : tuple
        Positional arguments to check against the signature.
    **kwargs : dict
        Keyword arguments to check against the signature.

    Returns
    -------
    BoundArguments or None
        The bound signature if the arguments match the signature, None otherwise.

    Examples
    --------
    >>> check_signature([('arg1', int), ('arg2', str)], 10, 'test')
    True
    >>> check_signature([('arg1', int), ('arg2', str)], 'test', 10)
    False
    >>> check_signature([('arg1', int)], arg1=5)
    True
    >>> check_signature([('arg1', int)], arg1='not an int')
    False
    """
    sig, allow_kwargs = build_signature(signature_spec)
    try:
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()
    except TypeError:
        return None

    for name, val in bound.arguments.items():
        expected = sig.parameters[name].annotation
        if expected is not inspect.Parameter.empty and not isinstance(val, expected):
            return None

    if not allow_kwargs and bound.kwargs:
        return None

    return bound


def build_signature(signature_spec: list):
    """Build a signature from a list of tokens.

    Parameters
    ----------
    signature_spec : list
        A list of tokens that define the function signature.
        Tokens can be strings or tuples with varying formats as detailed below.

    Returns
    -------
    inspect.Signature
        A Signature object representing the constructed function signature.
    """
    params = []
    seen_star = False

    for token in signature_spec:
        if token == "/":
            kind = inspect.Parameter.POSITIONAL_ONLY

        elif token == "*":
            seen_star = True
            continue  # next params will be KEYWORD_ONLY

        elif token == "**":
            # Anonymous **kwargs
            param = inspect.Parameter("kwargs", inspect.Parameter.VAR_KEYWORD)
            params.append(param)

        elif isinstance(token, str) and token.startswith("**"):
            name = token[2:]
            param = inspect.Parameter(name, inspect.Parameter.VAR_KEYWORD)
            params.append(param)

        elif isinstance(token, str) and token.startswith("*"):
            name = token[1:]
            seen_star = True
            param = inspect.Parameter(name, inspect.Parameter.VAR_POSITIONAL)
            params.append(param)

        elif isinstance(token, tuple):
            if not (1 <= len(token) <= 3):
                raise ValueError(f"Invalid tuple token: {token}")

            name = token[0]
            annotation = inspect.Parameter.empty
            default = inspect.Parameter.empty

            if len(token) == 2:
                # Annotation or default?
                if isinstance(token[1], type):
                    annotation = token[1]
                else:
                    default = token[1]
            elif len(token) == 3:
                annotation, default = token[1], token[2]

            kind = (
                inspect.Parameter.KEYWORD_ONLY
                if seen_star
                else inspect.Parameter.POSITIONAL_OR_KEYWORD
            )

            param = inspect.Parameter(
                name, kind, annotation=annotation, default=default
            )
            params.append(param)

        else:
            raise ValueError(f"Invalid token: {token}")

    allow_kwargs = any(
        token == "**" or (isinstance(token, str) and token.startswith("**"))
        for token in signature_spec
    )
    return inspect.Signature(params), allow_kwargs


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import pytest
import torch
import torch.nn as nn
from torchlinops import NUFFT, Dense, Diagonal, Dim, Stack, split_linop
from torchlinops.utils import memory_aware_to, MemReporter, tensor_memory_span

PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


class SampleModule(nn.Module):
    def __init__(self):
        super().__init__()
        base = torch.arange(4 * 64 * 64).float()
        self.shared_view = nn.Parameter(
            base.as_strided((2, 1, 64, 64), (4096, 4096, 64, 1), 8192)
        )
        self.same_view = nn.Parameter(self.shared_view)
        self.empty = nn.Parameter(torch.tensor([], dtype=torch.float32))
        self.zero_shape = nn.Parameter(torch.empty((0, 3, 224)))
        self.noncontig = nn.Parameter(
            torch.arange(12).view(3, 4).t(), requires_grad=False
        )  # transposed (non-contig)


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_model_storage_after_movement():
    model = SampleModule()
    model_cuda = memory_aware_to(model, torch.device("cuda"))
    assert (
        model_cuda.shared_view.untyped_storage()._cdata
        == model_cuda.same_view.untyped_storage()._cdata
    )
    assert model_cuda.empty.numel() == 0 and model_cuda.empty.device.type == "cuda"
    assert model_cuda.zero_shape.shape == (0, 3, 224)
    assert not model_cuda.noncontig.is_contiguous()

    assert (model.shared_view == model_cuda.shared_view).all()


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_model_storage_duplicate_submodules():
    weight = torch.randn(3, 4)
    P = Dense(weight, weightshape=Dim("MN"), ishape=Dim("N"), oshape=Dim("M"))
    A = Stack(P, P, P, odim_and_idx=("B", 0))
    memory_aware_to(A, torch.device("cuda"))
    assert A[0].weight.is_cuda
    assert id(A[0].weight) == id(A[1].weight)


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_model_storage_parameterlist():
    class ModuleWithParameterList(nn.Module):
        def __init__(self, *tensor_list):
            super().__init__()
            self.weights = nn.ParameterList(tensor_list)

    class ModuleWithDuplicateSubmodules(nn.Module):
        def __init__(self, module):
            super().__init__()
            self.modules = nn.ModuleList([module, module])

    base_tensor = torch.randn(6, 3)
    modulea = ModuleWithParameterList(
        base_tensor[..., 0], base_tensor[..., 1], base_tensor[..., 2]
    )
    moduleb = ModuleWithDuplicateSubmodules(modulea)
    memory_aware_to(moduleb, torch.device("cpu"))


def make_linop(trj, dcf, mps, nufft_width, nufft_oversamp, nufft_mode):
    """Sample linop with some complexity reflecting real-world use"""
    im_size = mps.shape[1:]
    P = trj.shape[0]
    linops = []
    S = Dense(
        mps,
        weightshape=Dim("CNxNyNz"),
        ishape=Dim("NxNyNz"),
        oshape=Dim("CNxNyNz"),
    )
    for p in range(P):
        # DCF
        Dp = Diagonal(dcf[p], ioshape=Dim("CTK"), broadcast_dims=Dim("C"))
        Fp = NUFFT(
            trj[p],
            im_size,
            output_shape=Dim("TK"),
            oversamp=nufft_oversamp,
            mode=nufft_mode,
        )
        Ap = (Dp ** (1 / 2)) @ Fp @ S
        linops.append(Ap)
    A = Stack(*linops, idim_and_idx=("P", 0), odim_and_idx=("P", 0))
    return A


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_full_linop():
    C, Nx, Ny, Nz = 5, 8, 8, 8
    T, K = 7, 11
    trj = Nx * torch.rand(T, K, 3) - Nx // 2
    trj = torch.round(trj)
    dcf = torch.randn(T, K)
    mps = torch.randn(C, Nx, Ny, Nz, dtype=torch.complex64)

    A = make_linop(
        trj,
        dcf,
        mps,
        nufft_width=4,
        nufft_oversamp=1.25,
        nufft_mode="sampling",
    )
    memory_aware_to(A, torch.device("cuda"))
    assert True


def check_span(t):
    min_off, max_off = tensor_memory_span(t)
    flat = t._base.view(-1) if t._base is not None else t.view(-1)
    used = t.clone().contiguous().view(-1)
    span = flat[min_off : max_off + 1]
    # Ensure the span includes the tensor's flattened values
    for v in used:
        assert (span == v).any(), f"Value {v.item()} not found in span"


@pytest.mark.parametrize("shape", [(10,), (4, 5), (2, 3, 4)])
def test_contiguous_tensor(shape):
    t = torch.arange(torch.tensor(shape).prod()).reshape(shape)
    min_off, max_off = tensor_memory_span(t)
    assert min_off == t.storage_offset()
    assert max_off == t.storage_offset() + t.numel() - 1
    check_span(t)


def test_strided_tensor():
    t = torch.arange(20).reshape(4, 5)
    v = t[::2, ::2]  # non-contiguous
    min_off, max_off = tensor_memory_span(v)
    check_span(v)


def test_negative_stride_1d():
    t = torch.arange(10)
    # v = t[::-1]
    v = t.flip(0)
    min_off, max_off = tensor_memory_span(v)
    assert min_off == 0
    assert max_off == 9
    check_span(v)


def test_negative_stride_2d():
    t = torch.arange(16).reshape(4, 4)
    # v = t[::-1, ::-1]
    v = t.flip(0, 1)
    min_off, max_off = tensor_memory_span(v)
    assert min_off == 0
    assert max_off == 15
    check_span(v)


def test_partial_negative_stride():
    t = torch.arange(24).reshape(2, 3, 4)
    # v = t[:, ::-1, ::2]  # only flip dim 1
    v = t.flip(1)[..., ::2]
    min_off, max_off = tensor_memory_span(v)
    check_span(v)


def test_storage_offset():
    t = torch.arange(100)
    v = t[10:20]
    min_off, max_off = tensor_memory_span(v)
    assert min_off == v.storage_offset()
    assert max_off == v.storage_offset() + v.numel() - 1
    check_span(v)


def test_subview_of_view():
    t = torch.arange(100)
    v1 = t[10:50]
    v2 = v1[::2]  # subview of a view
    min_off, max_off = tensor_memory_span(v2)
    check_span(v2)


def test_nontrivial_stride():
    t = torch.arange(3 * 5 * 7).reshape(3, 5, 7)
    # v = t[::2, ::-1, ::3]
    v = t.flip(1)[::2, :, ::3]
    min_off, max_off = tensor_memory_span(v)
    check_span(v)


def test_scalar_tensor():
    t = torch.tensor(42)
    min_off, max_off = tensor_memory_span(t)
    assert min_off == t.storage_offset()
    assert max_off == t.storage_offset()


def test_reconstruct_from_span_no_negative_stride():
    base = torch.arange(60).reshape(3, 4, 5)
    # Slice with positive strides only
    t = base[1::2, 1:3, ::2]  # shape (1, 2, 3), all strides positive

    min_off, max_off = tensor_memory_span(t)
    flat = base.view(-1)
    block = flat[min_off : max_off + 1].clone()

    # Compute offset relative to min_off (no adjustment needed for positive strides)
    new_offset = t.storage_offset() - min_off

    reconstructed = torch.as_strided(
        block,
        size=t.size(),
        stride=t.stride(),
        storage_offset=new_offset,
    )

    # Check that reconstruction matches original tensor exactly
    assert torch.equal(reconstructed, t), "Reconstructed tensor does not match original"


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_split_movement():
    """Messing around with block sizes"""
    M, N = 100, 10
    weight = torch.randn(M, N)

    # weight_half = weight[:, :5]
    # weight_half = weight_half.to(torch.device("cuda"))
    # breakpoint()

    # weight = weight.to(torch.device("cuda"))
    # breakpoint()

    A = Dense(weight, weightshape=Dim("MN"), ishape=Dim("N"), oshape=Dim("M"))

    # linops, _, _ = split_linop(A, {"N": N // 2})
    linops, _, _ = split_linop(A, {"M": M // 2})
    linop_gpu = memory_aware_to(linops[1], torch.device("cuda"))
    # linop_gpu_2 = linops[1].to(torch.device("cuda"))
    linop_cpu = linops[0]

    MemReporter().report(linop_gpu)
    MemReporter().report(linop_cpu)
